# distutils: language = c++

from libcpp.vector cimport vector
from libcpp cimport bool

import numpy as np
cimport numpy as np

from libc.math cimport exp, fabs, ceil

import ANNarchy
from ANNarchy.core.Random import RandomDistribution
from ANNarchy.core.Population import Population
from ANNarchy.intern.ConfigManagement import ConfigManager

cimport ANNarchy.cython_ext.Coordinates as Coordinates

def get_dt(obj):
    """
    Retrieve the *dt* configuration.

    Note:   the net_id is not directly available within the cython module.
            Therefore, we need to use the population object as interim step.
    """

    net_id = obj.net_id if isinstance(obj, Population) else obj.population.net_id

    return ConfigManager().get('dt', net_id=net_id)

##################################################
### Connector methods, these functions are    ####
### exported towards ConnectorMethods         ####
##################################################
def all_to_all(pre, post, weights, delays, allow_self_connections, storage_format, storage_order):
    """ Cython implementation of the all-to-all pattern."""
    # instantiate pattern
    projection = LILConnectivity(dt=get_dt(pre))
    projection.all_to_all(pre, post, weights, delays, allow_self_connections)

    return projection

def one_to_one(pre, post, weights, delays, storage_format, storage_order):
    """ Cython implementation of the one-to-one pattern."""
    # instantiate pattern
    projection = LILConnectivity(dt=get_dt(pre))
    projection.one_to_one(pre, post, weights, delays)

    return projection

def fixed_probability(pre, post, probability, weights, delays, allow_self_connections, storage_format, storage_order):
    """ Cython implementation of the fixed_probability pattern."""
    # instantiate pattern
    projection = LILConnectivity(dt=get_dt(pre))
    projection.fixed_probability(pre, post, probability, weights, delays, allow_self_connections)

    return projection

def fixed_number_pre(pre, post, int number, weights, delays, allow_self_connections, storage_format, storage_order):
    """ Cython implementation of the fixed_number_pre pattern."""
    # instantiate pattern
    projection = LILConnectivity(dt=get_dt(pre))
    projection.fixed_number_pre(pre, post, number, weights, delays, allow_self_connections)

    return projection

def fixed_number_post(pre, post, int number, weights, delays, allow_self_connections, storage_format, storage_order):
    """ Cython implementation of the fixed_number_post pattern."""
    # instantiate pattern
    projection = LILConnectivity(dt=get_dt(pre))
    projection.fixed_number_post(pre, post, number, weights, delays, allow_self_connections)

    return projection

def gaussian(pre_pop, post_pop, float amp, float sigma, delays, limit, allow_self_connections, storage_format, storage_order):
    """ Cython implementation of the gaussian pattern."""
    # instantiate pattern
    projection = LILConnectivity(dt=get_dt(pre_pop))
    projection.gaussian(pre_pop, post_pop, amp, sigma, delays, limit, allow_self_connections)

    return projection

def dog(pre_pop, post_pop, float amp_pos, float sigma_pos, float amp_neg, float sigma_neg, delays, limit, allow_self_connections, storage_format, storage_order):
    """ Cython implementation of the difference-of-gaussian (dog) pattern."""
    # instantiate pattern
    projection = LILConnectivity(dt=get_dt(pre_pop))
    projection.dog(pre_pop, post_pop, amp_pos, sigma_pos, amp_neg, sigma_neg, delays, limit, allow_self_connections)

    return projection

###################################################
########## LIL object to hold synapses ############
###################################################
cdef class LILConnectivity:

    def __cinit__(self, dt):
        self.max_delay = 0
        self.size = 0
        self.nb_synapses = 0
        self.uniform_delay = -1
        self.dt = dt
        self.requires_sorting = False
        self.last_added_idx = -1

    def __dealloc__(self):
        self.post_rank.clear()
        self.post_rank.shrink_to_fit()
        self.pre_rank.clear()
        self.pre_rank.shrink_to_fit()
        self.w.clear()
        self.w.shrink_to_fit()
        self.delay.clear()
        self.delay.shrink_to_fit()
        # not sure, if really needed ...
        self.requires_sorting = False
        self.last_added_idx = -1

    cpdef add(self, int rk, r, w, d):
        self.push_back(rk, r, w, d)

    cpdef push_back(self, int rk, vector[int] r, vector[double] w, vector[double] d):
        cdef unsigned int i
        cdef vector[int] int_delays
        cdef int max_d, unif_d

        # sanity check: added rows should be ascending sorted
        if rk < self.last_added_idx:
            if self.requires_sorting == False:
                ANNarchy.intern.Messages._warning("LILConnectivity.add()/.push_back(): dendrites should be added in an ascending order for performance reasons.")
                ANNarchy.intern.Messages._print("ANNarchy will sort the dendrites during compile() which increases the required time.")
            self.requires_sorting = True
        else:
            self.last_added_idx = rk

        # Do not add empty arrays
        if r.size() == 0:
            return

        # Store the connectivity
        self.post_rank.push_back(rk)
        self.pre_rank.push_back(r)

        # Are the weights uniform?
        if w.size() > 1 or r.size() == 1:
            self.w.push_back(w)
        else:
            self.w.push_back(vector[double](r.size(), w[0]))

        # Store delays and update the max
        for i in range(d.size()):
            int_delays.push_back(round(d[i]/self.dt))
        self.delay.push_back(int_delays)
        max_d = round(np.max(d)/self.dt)
        if max_d > self.max_delay:
            self.max_delay = max_d

        # Are the delays uniform?
        if d.size() > 1 :
            self.uniform_delay = -1
        else:
            unif_d = round(d[0]/self.dt)
            if self.uniform_delay != unif_d and self.size > 0:
                self.uniform_delay = -1
            else:
                self.uniform_delay = unif_d

        # Increase the size
        self.size += 1
        self.nb_synapses += r.size()

    cpdef int get_max_delay(self):
        return self.max_delay

    cpdef int get_uniform_delay(self):
        return self.uniform_delay

    cpdef compute_average_row_length(self):
        cdef vector[int] rl
        for i in range(self.pre_rank.size()):
            rl.push_back(self.pre_rank[i].size())

        return np.mean(rl), np.std(rl), np.amin(rl), np.amax(rl)

    cpdef compute_average_col_idx_gap(self):
        cdef vector[int] idx_gap

        for i in range(self.pre_rank.size()):
            for j in range(self.pre_rank[i].size()-1):
                idx_gap.push_back(self.pre_rank[i][j+1]-self.pre_rank[i][j])

        return np.mean(idx_gap), np.std(idx_gap)

    cpdef validate(self):
        cdef int idx, single, rk
        cdef vector[int] ranks
        cdef vector[double] weights
        cdef vector[int] delays
        cdef dict doubletons = {}
        cdef list postranks = list(self.post_rank)
        cdef list set_postranks = list(set(postranks))
        cdef list preranks, set_preranks, indices

        if len(postranks) != len(set_postranks):
            ANNarchy.intern.Messages._warning('You have added several times the same post-synaptic neuron to the LIL data in your connector method.')
            ANNarchy.intern.Messages._print('ANNarchy will try to sort the entries if possible, it may take some time...')
        else:
            return

        # Find out which post neurons are doubled.
        for single in set_postranks:
            doubletons[single] = []
            for idx, possible_double in enumerate(postranks):
                if possible_double == single:
                    doubletons[single].append(idx)
            if len(doubletons[single]) == 1 : # Only one occurence
                doubletons.pop(single)

        # Gather the info
        for rk, indices in doubletons.items():
            # Store the different infos
            ranks.clear()
            weights.clear()
            delays.clear()
            for idx in reversed(indices): # Start from below, otherwise deletion crashes
                # Gather data
                ranks.insert(ranks.end(), self.pre_rank[idx].begin(), self.pre_rank[idx].end())
                weights.insert(weights.end(), self.w[idx].begin(), self.w[idx].end())
                delays.insert(delays.end(), self.delay[idx].begin(), self.delay[idx].end())
                # Delete the old vectors
                self.post_rank.erase(self.post_rank.begin()+idx)
                self.pre_rank.erase(self.pre_rank.begin()+idx)
                self.w.erase(self.w.begin()+idx)
                self.delay.erase(self.delay.begin()+idx)

            # Check if no synapse is doubled
            preranks = list(ranks)
            set_preranks = list(set(preranks))
            if len(preranks) != len(set_preranks):
                ANNarchy.intern.Messages._error('The same synapse has been declared multiple times! Check your code.', exit=True)

            # Add the new data
            self.post_rank.push_back(rk)
            self.pre_rank.push_back(ranks)
            self.w.push_back(weights)
            self.delay.push_back(delays)

    #####################################################
    # Connector method implementations for list-of-list #
    #####################################################
    cpdef all_to_all(self, pre, post, weights, delays, allow_self_connections):
        " Implementation of the all-to-all pattern "
        cdef double weight
        cdef int r_post, size_pre, i
        cdef list tmp, post_ranks, pre_ranks
        cdef vector[int] r
        cdef vector[double] w, d

        # Retríeve ranks
        post_ranks = post.ranks.tolist()
        pre_ranks = pre.ranks.tolist()

        for r_post in post_ranks:
            # List of pre ranks
            tmp = [i for i  in pre_ranks]
            if not allow_self_connections:
                try:
                    tmp.remove(r_post)
                except: # was not in the list
                    pass
            r = tmp
            size_pre = len(tmp)
            # Weights
            if isinstance(weights, (int, float)):
                weight = weights
                w = vector[double](1, weight)
            elif isinstance(weights, RandomDistribution):
                w = weights.get_list_values(size_pre)
            # Delays
            if isinstance(delays, (float, int)):
                d = vector[double](1, delays)
            elif isinstance(delays, RandomDistribution):
                d = delays.get_list_values(size_pre)

            # Create the dendrite
            self.push_back(r_post, r, w, d)

    cpdef one_to_one(self, pre, post, weights, delays):
        """ Cython implementation of the one-to-one pattern."""
        cdef int idx
        cdef list post_ranks, pre_ranks
        cdef vector[int] r
        cdef vector[double] w, d

        # Retríeve ranks
        post_ranks = post.ranks.tolist()
        pre_ranks = pre.ranks.tolist()
    
        for idx in range(len(post_ranks)):
            if idx >= pre.size:
                break
            r = vector[int](1, pre_ranks[idx])
            # Get the weights and delays
            w, d = _get_weights_delays(1, weights, delays)
            # Create the dendrite
            self.push_back(post_ranks[idx], r, w, d)

    cpdef fixed_probability(self, pre, post, probability, weights, delays, allow_self_connections):
        " Implementation of the fixed-probability pattern "
        cdef double weight
        cdef int r_post, r_pre, size_pre, max_size_pre
        cdef list post_ranks
        cdef vector[int] r
        cdef vector[double] w, d
        cdef np.ndarray random_values, tmp, pre_ranks

        # Retríeve ranks
        post_ranks = post.ranks.tolist()
        pre_ranks = pre.ranks
        max_size_pre = len(pre.ranks)

        for r_post in post_ranks:
            # List of pre ranks
            random_values = np.random.random(max_size_pre)
            tmp = pre_ranks[random_values < probability]
            if not allow_self_connections:
                tmp = tmp[tmp != r_post]

            # sort the indices to prevent irregular accesses
            r = np.sort(tmp)
            size_pre = tmp.size
            if size_pre == 0:
                continue

            # Weights
            if isinstance(weights, (int, float)):
                weight = weights
                w = vector[double](1, weight)
            elif isinstance(weights, RandomDistribution):
                w = weights.get_list_values(size_pre)
            # Delays
            if isinstance(delays, (float, int)):
                d = vector[double](1, delays)
            elif isinstance(delays, RandomDistribution):
                d = delays.get_list_values(size_pre)
            # Create the dendrite
            self.push_back(r_post, r, w, d)

    cpdef fixed_number_pre(self, pre, post, int number, weights, delays, allow_self_connections):
        cdef double weight
        cdef int r_post, r_pre, size_pre
        cdef list pre_ranks, post_ranks
        cdef vector[int] r
        cdef vector[double] w, d

        # Retríeve ranks
        post_ranks = post.ranks.tolist()
        pre_ranks = pre.ranks.tolist()

        for r_post in post_ranks:
            # List of pre ranks
            r = np.random.choice(pre_ranks, size=number, replace=False)
            if len(r) == 0:
                continue
            if not allow_self_connections:
                while r_post in list(r): # the post index is in the list
                    r = np.random.choice(pre_ranks, size=number, replace=False)

            # sort the indices to prevent irregular accesses
            r = np.sort(r)

            # Weights
            if isinstance(weights, (int, float)):
                weight = weights
                w = vector[double](1, weight)
            elif isinstance(weights, RandomDistribution):
                w = weights.get_list_values(number)
            # Delays
            if isinstance(delays, (float, int)):
                d = vector[double](1, delays)
            elif isinstance(delays, RandomDistribution):
                d = delays.get_list_values(number)
            # Create the dendrite
            self.push_back(r_post, r, w, d)

    cpdef fixed_number_post(self, pre, post, int number, weights, delays, allow_self_connections):
        cdef double weight
        cdef int r_post, r_pre, size_pre
        cdef list pre_ranks, post_ranks
        cdef list pre_r
        cdef dict rk_mat
        cdef vector[int] r
        cdef vector[int] tmp
        cdef vector[double] w, d

        # Retrieve ranks
        post_ranks = post.ranks.tolist()
        pre_ranks = pre.ranks.tolist()

        # Build the backward matrix
        rk_mat = {i: [] for i in post_ranks}
        for r_pre in pre_ranks:
            if number >= len(post_ranks):
                tmp = post_ranks
            else:
                tmp = np.random.choice(post_ranks, size=number, replace=False)
                if not allow_self_connections:
                    # the post index is in the list, redraw
                    while r_pre in list(tmp):   # TODO: maybe a find() would be better
                        tmp = np.random.choice(post_ranks, size=number, replace=False)
            for i in tmp:
                rk_mat[i].append(r_pre)

        # Create the dendrites
        for r_post in post_ranks:
            # List of pre ranks
            r = rk_mat[r_post]
            size_pre = len(r)
            if size_pre == 0:
                continue
            # Weights
            if isinstance(weights, (int, float)):
                weight = weights
                w = vector[double](1, weight)
            elif isinstance(weights, RandomDistribution):
                w = weights.get_list_values(size_pre)
            # Delays
            if isinstance(delays, (float, int)):
                d = vector[double](1, delays)
            elif isinstance(delays, RandomDistribution):
                d = delays.get_list_values(size_pre)
            # Create the dendrite
            self.push_back(r_post, r, w, d)

    cpdef gaussian(self, pre_pop, post_pop, float amp, float sigma, delays, limit, allow_self_connections):
        cdef float distance, value
        cdef int post, pre, pre_size, post_size, c, nb_synapses, pre_dim, post_dim
        cdef tuple pre_coord, post_coord
        cdef tuple pre_geometry, post_geometry
        cdef list ranks, values

        cdef vector[int] r
        cdef vector[double] w, d

        # Population sizes
        pre_geometry = pre_pop.geometry
        post_geometry = post_pop.geometry
        if isinstance(pre_geometry, int):
            pre_size = pre_geometry
            pre_dim = 1
        else:
            pre_dim = len(pre_geometry)
            pre_size = 1
            for c in pre_geometry:
                pre_size  = pre_size * c
        if isinstance(post_geometry, int):
            post_size = post_geometry
            post_dim = 1
        else:
            post_dim = len(post_geometry)
            post_size = 1
            for c in post_geometry:
                post_size  = post_size * c

        # Create the projection data
        for post in list(range(post_size)):
            ranks = []
            values = []
            if post_dim == 1:
                post_coord = (post/float(post_size-1), )
            elif post_dim == 2:
                post_coord = Coordinates.get_normalized_2d_coord(post, post_geometry)
            elif post_dim == 3:
                post_coord = Coordinates.get_normalized_3d_coord(post, post_geometry)
            else:
                post_coord = Coordinates.get_normalized_coord(post, post_geometry)
            for pre in list(range(pre_size)):
                if not allow_self_connections and pre==post:
                    continue
                if pre_dim == 1:
                    pre_coord = (pre/float(pre_size-1), )
                    distance = Coordinates.comp_dist1D(pre_coord, post_coord)
                elif pre_dim == 2:
                    pre_coord = Coordinates.get_normalized_2d_coord(pre, pre_geometry)
                    distance = Coordinates.comp_dist2D(pre_coord, post_coord)
                elif pre_dim == 3:
                    pre_coord = Coordinates.get_normalized_3d_coord(pre, pre_geometry)
                    distance = Coordinates.comp_dist3D(pre_coord, post_coord)
                else:
                    pre_coord = Coordinates.get_normalized_coord(pre, pre_geometry)
                    distance = Coordinates.comp_distND(pre_coord, post_coord)
                value = amp * exp(-distance/(2.0*sigma**2))
                if value > limit * amp:
                    ranks.append(pre)
                    values.append(value)
            nb_synapses = len(ranks)
            r = ranks
            w = values
            if isinstance(delays, (float, int)):
                d = vector[double](1, delays)
            elif isinstance(delays, RandomDistribution):
                d = delays.get_list_values(nb_synapses)
            # Create the dendrite
            self.push_back(post, r, w, d)

    cpdef dog(self, pre_pop, post_pop, float amp_pos, float sigma_pos, float amp_neg, float sigma_neg, delays, limit, allow_self_connections):
        cdef float distance, value
        cdef int post, pre, pre_size, post_size, c, nb_synapses, pre_dim, post_dim
        cdef tuple pre_coord, post_coord
        cdef tuple pre_geometry, post_geometry
        cdef list ranks, values

        cdef vector[int] r
        cdef vector[double] w, d

        # Population sizes
        pre_geometry = pre_pop.geometry
        post_geometry = post_pop.geometry
        if isinstance(pre_geometry, int):
            pre_size = pre_geometry
            pre_dim = 1
        else:
            pre_dim = len(pre_geometry)
            pre_size = 1
            for c in pre_geometry:
                pre_size  = pre_size * c
        if isinstance(post_geometry, int):
            post_size = post_geometry
            post_dim = 1
        else:
            post_dim = len(post_geometry)
            post_size = 1
            for c in post_geometry:
                post_size  = post_size * c

        # Create the projection data as LIL
        for post in list(range(post_size)):
            ranks = []
            values = []
            if post_dim == 1:
                post_coord = (post/float(post_size-1), )
            elif post_dim == 2:
                post_coord = Coordinates.get_normalized_2d_coord(post, post_geometry)
            elif post_dim == 3:
                post_coord = Coordinates.get_normalized_3d_coord(post, post_geometry)
            else:
                post_coord = Coordinates.get_normalized_coord(post, post_geometry)
            for pre in list(range(pre_size)):
                if not allow_self_connections and pre==post:
                    continue
                if pre_dim == 1:
                    pre_coord = (pre/float(pre_size-1), )
                    distance = Coordinates.comp_dist1D(pre_coord, post_coord)
                elif pre_dim == 2:
                    pre_coord = Coordinates.get_normalized_2d_coord(pre, pre_geometry)
                    distance = Coordinates.comp_dist2D(pre_coord, post_coord)
                elif pre_dim == 3:
                    pre_coord = Coordinates.get_normalized_3d_coord(pre, pre_geometry)
                    distance = Coordinates.comp_dist3D(pre_coord, post_coord)
                else:
                    pre_coord = Coordinates.get_normalized_coord(pre, pre_geometry)
                    distance = Coordinates.comp_distND(pre_coord, post_coord)
                value = amp_pos * exp(-distance/(2.0*sigma_pos**2)) - amp_neg * exp(-distance/(2.0*sigma_neg**2))
                if fabs(value) > limit * fabs(amp_pos - amp_neg):
                    ranks.append(pre)
                    values.append(value)
            nb_synapses = len(ranks)
            r = ranks
            w = values
            if isinstance(delays, (float, int)):
                d = vector[double](1, delays)
            elif isinstance(delays, RandomDistribution):
                d = delays.get_list_values(nb_synapses)

            # Create the dendrite
            self.push_back(post, r, w, d)

cdef _get_weights_delays(int size, weights, delays):

    cdef vector[double] w, d
    cdef list tmp

    # Weights
    if isinstance(weights, (int, float)):
        w = vector[double](1, weights)
    elif isinstance(weights, RandomDistribution):
        tmp = weights.get_list_values(size)
        w = tmp
    # Delays
    if isinstance(delays, (float, int)):
        d = vector[double](1, delays)
    elif isinstance(delays, RandomDistribution):
        d = delays.get_list_values(size)

    return w, d



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

convolve_template_omp = {
    # Declare the connectivity matrix
    'declare_connectivity_matrix': """
    // Connectivity data
    std::vector< std::vector<int> > pre_coords;
    """,

    # Accessors for the connectivity matrix
    'access_connectivity_matrix': """
    // Accessor to connectivity data
    std::vector<int> get_post_ranks() { 
        std::vector<int> v(pre_coords.size());
        std::iota (std::begin(v), std::end(v), 0);
        return v;
    }
""" ,

    # Export the connectivity matrix
    'export_connectivity': "",

    # Variables for the psp code
    'psp_prefix': """
        int rk_pre;
        %(float_prec)s sum=0.0;
""",

    # Nanobind
    'wrapper': """
    // Convolution ProjStruct%(id_proj)s
    nanobind::class_<ProjStruct%(id_proj)s>(m, "proj%(id_proj)s_wrapper")
        // Constructor
        .def(nanobind::init<>())

        // Flags
        .def_rw("_transmission", &ProjStruct%(id_proj)s::_transmission)
        .def_rw("_axon_transmission", &ProjStruct%(id_proj)s::_axon_transmission)
        .def_rw("_update", &ProjStruct%(id_proj)s::_update)
        .def_rw("_plasticity", &ProjStruct%(id_proj)s::_plasticity)

        // Connectivity
        .def("post_rank", nanobind::overload_cast<>(&ProjStruct%(id_proj)s::get_post_ranks))
        .def_rw("w", &ProjStruct%(id_proj)s::w)
        .def_rw("pre_coords", &ProjStruct%(id_proj)s::pre_coords)
        %(delays)s

        // Other methods
        .def("clear", &ProjStruct%(id_proj)s::clear);
""",


    # Memory Management
    'clear': """
        // pre-coords
        for (auto it = pre_coords.begin(); it != pre_coords.end(); it++) {
            it->clear();
            it->shrink_to_fit();
        }
        pre_coords.clear();
        pre_coords.shrink_to_fit();
""",

    'size_in_bytes': """
        // pre-coords
        size_in_bytes += sizeof(std::vector<std::vector<int>>);
        size_in_bytes += pre_coords.capacity() * sizeof(std::vector<int>);
        for (auto it = pre_coords.begin(); it != pre_coords.end(); it++) {
            size_in_bytes += it->capacity() * sizeof(int);
        }
"""
}

convolve_template_cuda = {
    'include_additional': '#include "VecTransformation.hpp"',

    # Declare the connectivity matrix
    'declare_connectivity_matrix': """
    // Connectivity data
    std::vector< std::vector<int> > pre_coords;
    int* gpu_pre_coords = nullptr;
    bool pre_coords_dirty = false;
    """,

    # Accessors for the connectivity matrix
    'access_connectivity_matrix': """
    // Accessor to connectivity data
    std::vector<int> get_post_ranks() { 
        std::vector<int> v(pre_coords.size());
        std::iota (std::begin(v), std::end(v), 0);
        return v;
    }    
""" ,

    # Export the connectivity matrix
    'export_connectivity': "",

    # Variables for the psp code
    'psp_prefix': """
        int rk_pre;
        %(float_prec)s sum=0.0;
""",

    # Nanobind
    'wrapper': """
    // Convolution ProjStruct%(id_proj)s
    nanobind::class_<ProjStruct%(id_proj)s>(m, "proj%(id_proj)s_wrapper")
        // Constructor
        .def(nanobind::init<>())

        // Flags
        .def_rw("_transmission", &ProjStruct%(id_proj)s::_transmission)
        .def_rw("_axon_transmission", &ProjStruct%(id_proj)s::_axon_transmission)
        .def_rw("_update", &ProjStruct%(id_proj)s::_update)
        .def_rw("_plasticity", &ProjStruct%(id_proj)s::_plasticity)

        // Connectivity
        .def("post_rank", nanobind::overload_cast<>(&ProjStruct%(id_proj)s::get_post_ranks))
        .def_rw("pre_coords", &ProjStruct%(id_proj)s::pre_coords)
        .def_rw("pre_coords_dirty", &ProjStruct%(id_proj)s::pre_coords_dirty)
        .def_rw("w", &ProjStruct%(id_proj)s::w)
        .def_rw("w_dirty", &ProjStruct%(id_proj)s::host_w_dirty)

        // Other methods
        .def("clear", &ProjStruct%(id_proj)s::clear);
""",

    # Memory Management
    'clear': """
        // pre-coords
        for (auto it = pre_coords.begin(); it != pre_coords.end(); it++) {
            it->clear();
            it->shrink_to_fit();
        }
        pre_coords.clear();
        pre_coords.shrink_to_fit();

        cudaFree(gpu_pre_coords);
""",

    'size_in_bytes': """
        // pre-coords
        size_in_bytes += sizeof(std::vector<std::vector<int>>);
        size_in_bytes += pre_coords.capacity() * sizeof(std::vector<int>);
        for (auto it = pre_coords.begin(); it != pre_coords.end(); it++) {
            size_in_bytes += it->capacity() * sizeof(int);
        }
""",

    # This template concerns only the connectivity where
    # no read-back is required
    "host_device_transfer": """
        if (pre_coords_dirty) {
        #ifdef _DEBUG
            std::cout << "ProjStruct%(id_proj)s (convolution): update device coords." << std::endl;
        #endif
            auto flat_coords = transform_2d_to_1d<int>(pre_coords);
            size_t size_in_bytes = flat_coords.size() * sizeof(int);

            cudaMalloc((void**)&gpu_pre_coords, size_in_bytes);
            cudaMemcpy(gpu_pre_coords, flat_coords.data(), size_in_bytes, cudaMemcpyHostToDevice);
        }
""",
    "device_host_transfer": "",

}

conv_filter_template = {

    # Single-thread/OpenMP
    "openmp": {
        "access": """
    // Local parameter w
    %(type_w)s get_w() { return w; }
    void set_w(%(type_w)s value) { w = value; }
""",
    },

    # CUDA
    "cuda": {
        "declare": """\t// Filter definition
    %(cpu_side_filter)s
    %(float_prec)s *gpu_w;
    bool host_w_dirty;
""",
        "access": """
    // Local parameter w
    %(type_w)s get_w() { return w; }
    void set_w(%(type_w)s value) {
    #ifdef _DEBUG
        std::cout << "ProjStruct%(id_proj)s (convolution): set new filter on host" << std::endl;
    #endif
        w = value;
        host_w_dirty = true;
    }
""",
    'host_device_transfer': """
        if ( host_w_dirty ) {
        #ifdef _DEBUG
            std::cout << "ProjStruct%(id_proj)s (convolution): update device filter." << std::endl;
        #endif
            auto flat_data = transform_%(pre_dim)sd_to_1d<%(ctype)s>(w);
            auto size_in_bytes = flat_data.size() * sizeof(%(ctype)s);

            cudaMalloc((void**)&gpu_w, size_in_bytes);
            auto malloc_err = cudaGetLastError();
            if (malloc_err != cudaSuccess) {
                std::cerr << "ProjStruct%(id_proj)s::host_to_device - cudaMalloc: " << cudaGetErrorString(malloc_err) << std::endl;
            }

            cudaMemcpy(gpu_w, flat_data.data(), size_in_bytes, cudaMemcpyHostToDevice);
            auto memcpy_err = cudaGetLastError();
            if (memcpy_err != cudaSuccess) {
                std::cerr << "ProjStruct%(id_proj)s::host_to_device - cudaMemcpy: " << cudaGetErrorString(memcpy_err) << std::endl;
            }

            host_w_dirty = false;
        }
""",
    # No read-back needed, as weights are not changed during simulation
    'device_host_transfer': ""
    }
}

cuda_convolution_single_filter = {
    "body": """__global__ void cu_conv_proj%(id_proj)s(%(float_prec)s* __restrict__ psp, const int* __restrict__ pre_coords, const %(float_prec)s* __restrict__ w%(pre_variables_header)s) {
    int bIdx = blockIdx.x;
    %(float_prec)s sum;
    int rk_pre, w_idx;
    const int *coord = &pre_coords[%(pre_dim)s*bIdx];

%(convolve_code)s

    psp[bIdx] += sum;
}
""",
    "invoke": """
void convolution_proj%(id_proj)s(RunConfig cfg, %(float_prec)s* __restrict__ psp, const int* __restrict__ pre_coords, const %(float_prec)s* __restrict__ w%(pre_variables_header)s) {
    cu_conv_proj%(id_proj)s<<<cfg.nb, cfg.tpb, cfg.smem_size, cfg.stream>>>(psp, pre_coords, w%(pre_variables_invoke)s);
}
""",
    "header": "void convolution_proj%(id_proj)s(RunConfig cfg, %(float_prec)s* __restrict__ psp, const int* __restrict__ pre_coords, const %(float_prec)s* __restrict__ w%(pre_variables_header)s);",
    "call": """
    if (proj%(id_proj)s->_transmission && pop%(id_post)s->_active ) {
        convolution_proj%(id_proj)s(
            RunConfig(pop%(id_post)s->size, 1, 0, proj%(id_proj)s->stream),
            pop%(id_post)s->gpu__sum_%(target)s, proj%(id_proj)s->gpu_pre_coords, proj%(id_proj)s->gpu_w%(pre_variables_call)s
        );
    
    #ifdef _DEBUG
        auto proj%(id_proj)s_conv_err = cudaDeviceSynchronize();
        if ( proj%(id_proj)s_conv_err != cudaSuccess) {
            std::cout << "Convolution projection %(id_proj)s - psp: " << cudaGetErrorString( proj%(id_proj)s_conv_err ) << std::endl;
        }
    #endif
    }
"""
}

cuda_convolution_bank_of_filter = {
    "body": """__global__ void cu_conv_proj%(id_proj)s(%(float_prec)s* __restrict__ psp, const int* __restrict__ pre_coords, const %(float_prec)s* __restrict__ w%(pre_variables_header)s) {
    int bIdx = blockIdx.x;
    %(float_prec)s sum;
    int rk_pre, w_idx;
    const int *coord = &pre_coords[%(filter_dim)s*bIdx];
    const %(float_prec)s *w_bank = &w[coord[%(pre_dim)s] * %(num_elem_filter)s];

%(convolve_code)s

    psp[bIdx] += sum;
}
""",
    "invoke": """void convolution_proj%(id_proj)s(RunConfig cfg, %(float_prec)s* psp, const int* pre_coords, const %(float_prec)s* w%(pre_variables_header)s) {
    cu_conv_proj%(id_projs)s<<<cfg.nb, cfg.tpb, cfg.smem_size, cfg.stream(psp, pre_coords, w%(pre_variables_invoke)s);
}
""",
    "header": "void convolution_proj%(id_proj)s(RunConfig cfg, %(float_prec)s* psp, const int* pre_coords, const %(float_prec)s* w%(pre_variables_header)s);",
    "call": """
    if (proj%(id_proj)s->_transmission && pop%(id_post)s->_active ) {
        convolution_proj%(id_proj)s(
            RunConfig(pop%(id_post)s->size, 1, 0, proj%(id_proj)s->stream),
            pop%(id_post)s->gpu__sum_%(target)s, proj%(id_proj)s->gpu_pre_coords, proj%(id_proj)s->gpu_w%(pre_variables_call)s
        );

    #ifdef _DEBUG
        auto proj%(id_proj)s_conv_err = cudaDeviceSynchronize();
        if ( proj%(id_proj)s_conv_err != cudaSuccess) {
            std::cout << "Convolution projection %(id_proj)s - psp: " << cudaGetErrorString( proj%(id_proj)s_conv_err ) << std::endl;
        }
    #endif
    }
"""
}

# Specialized code template: 3D filter bank, 2D pre-population
cuda_convolution_bank_of_filter_3d = {
    "body": """__global__ void cu_conv_proj%(id_proj)s(%(float_prec)s* __restrict__ psp, const int* __restrict__ pre_coords, const %(float_prec)s* __restrict__ w%(pre_variables_header)s) {
    int bIdx = blockIdx.x;
    int i_w = threadIdx.x;
    int j_w = threadIdx.y;
    int w_idx = %(filter_dim_j)s*(i_w) + j_w;

    const int *coord = &pre_coords[3*bIdx];
    const double *w_bank = &w[coord[2] * %(num_elem_filter)s];

    int i_pre = coord[0] + (i_w - %(pre_offset_i)s);
    if ((i_pre < 0) || (i_pre > %(pre_border_i)s)) {
        return;
    }

    int j_pre = coord[1] + (j_w - %(pre_offset_j)s);
    if ((j_pre < 0) || (j_pre > %(pre_border_j)s)) {
        return;
    }

    int rk_pre = %(pre_dim_j)s*(i_pre) + j_pre;
    atomicAdd(&psp[bIdx], %(pre_variable)s[rk_pre]*w_bank[w_idx]);
}
""",
    "invoke": """
void convolution_proj%(id_proj)s(RunConfig cfg, %(float_prec)s* psp, const int* pre_coords, const %(float_prec)s* w%(pre_variables_header)s) {
    cu_conv_proj%(id_proj)s<<<cfg.nb, cfg.tpb, cfg.smem_size, cfg.stream>>>(psp, pre_coords, w%(pre_variables_invoke)s);
}
""",
    "header": "void convolution_proj%(id_proj)s(RunConfig cfg, %(float_prec)s* psp, const int* pre_coords, const %(float_prec)s* w%(pre_variables_header)s);",
    "call": """
    if (proj%(id_proj)s->_transmission && pop%(id_post)s->_active ) {
        auto num_blocks = dim3(pop%(id_post)s->size, 1, 1);
        auto thread_config = dim3(%(filter_dim_i)s, %(filter_dim_j)s, 1);
        convolution_proj%(id_proj)s(
            RunConfig(num_blocks, thread_config, 0, proj%(id_proj)s->stream),
            pop%(id_post)s->gpu__sum_%(target)s, proj%(id_proj)s->gpu_pre_coords, proj%(id_proj)s->gpu_w%(pre_variables_call)s
        );

    #ifdef _DEBUG
        auto proj%(id_proj)s_conv_err = cudaDeviceSynchronize();
        if ( proj%(id_proj)s_conv_err != cudaSuccess) {
            std::cout << "Convolution projection %(id_proj)s - psp: " << cudaGetErrorString( proj%(id_proj)s_conv_err ) << std::endl;
        }
    #endif
    }
"""
}

cuda_convolution_bank_of_filter_4d = {
    "body": """__global__ void cu_conv_proj%(id_proj)s(%(float_prec)s* __restrict__ psp, const int* __restrict__ pre_coords, const %(float_prec)s* __restrict__ w%(pre_variables_header)s) {
    int bIdx = blockIdx.x;
    int i_w = threadIdx.x;
    int j_w = threadIdx.y;
    int rk_pre, w_idx;

    const int *coord = &pre_coords[4*bIdx];
    const double *w_bank = &w[coord[3] * %(num_elem_filter)s];

    int i_pre = coord[0] + (i_w - %(pre_offset_i)s);
    if ((i_pre < 0) || (i_pre > %(pre_border_i)s)) {
        return;
    }

    int j_pre = coord[1] + (j_w - %(pre_offset_j)s);
    if ((j_pre < 0) || (j_pre > %(pre_border_j)s)) {
        return;
    }

    %(float_prec)s sum = 0.0;
    for (int k_w = 0; k_w < 64; k_w++) {
        int k_pre = coord[2] + (k_w - %(pre_offset_k)s);
        if ((k_pre < 0) || (k_pre > %(pre_border_k)s)) {
            continue;
        }

        rk_pre = %(pre_dim_k)s * ( %(pre_dim_j)s*(i_pre) + j_pre ) + k_pre;
        w_idx = %(filter_dim_k)s*(%(filter_dim_j)s*(i_w) + j_w) + k_w;
        sum += %(pre_variable)s[rk_pre]*w_bank[w_idx];
    }

    atomicAdd(&psp[bIdx], sum);
}
""",
    "invoke": """
void convolution_proj%(id_proj)s(RunConfig cfg, %(float_prec)s* psp, const int* pre_coords, const %(float_prec)s* w%(pre_variables_header)s) {
    cu_conv_proj%(id_proj)s<<<cfg.nb, cfg.tpb, cfg.smem_size, cfg.stream>>>(psp, pre_coords, w%(pre_variables_invoke)s);
}
""",
    "header": "void convolution_proj%(id_proj)s(RunConfig cfg, %(float_prec)s* psp, const int* pre_coords, const %(float_prec)s* w%(pre_variables_header)s);",
    "call": """
    if (proj%(id_proj)s->_transmission && pop%(id_post)s->_active ) {
        auto num_blocks = dim3(pop%(id_post)s->size, 1, 1);
        auto thread_config = dim3(%(filter_dim_i)s, %(filter_dim_j)s, 1);
        convolution_proj%(id_proj)s(
            RunConfig(num_blocks, thread_config, 0, proj%(id_proj)s->stream),
            pop%(id_post)s->gpu__sum_%(target)s, proj%(id_proj)s->gpu_pre_coords, proj%(id_proj)s->gpu_w%(pre_variables_call)s
        );

    #ifdef _DEBUG
        auto proj%(id_proj)s_conv_err = cudaDeviceSynchronize();
        if ( proj%(id_proj)s_conv_err != cudaSuccess) {
            std::cout << "Convolution projection %(id_proj)s - psp: " << cudaGetErrorString( proj%(id_proj)s_conv_err ) << std::endl;
        }
    #endif
    }
"""
}




"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

import numpy as np

from ANNarchy.intern.SpecificPopulation import SpecificPopulation
from ANNarchy.intern.ConfigManagement import ConfigManager
from ANNarchy.intern import Messages
from ANNarchy.core.Population import Population
from ANNarchy.core.Neuron import Neuron



class TimedPoissonPopulation(SpecificPopulation):
    """
    Poisson population whose rate vary with the provided schedule.

    Example:

    ```python
    inp = net.create(
        TimedPoissonPopulation(
            geometry = 100,
            rates = [10., 20., 100., 20., 5.],
            schedule = [0., 100., 200., 500., 600.],
        )
    )
    ```

    This creates a population of 100 Poisson neurons whose rate will be:
    
    * 10 Hz during the first 100 ms.
    * 20 HZ during the next 100 ms.
    * 100 Hz during the next 300 ms.
    * 20 Hz during the next 100 ms.
    * 5 Hz until the end of the simulation.
    
    
    If you want the TimedPoissonPopulation to "loop" over the schedule, you can specify a period:

    ```python
    inp = net.create(
        TimedPoissonPopulation(
            geometry = 100,
            rates = [10., 20., 100., 20., 5.],
            schedule = [0., 100., 200., 500., 600.],
            period = 1000.,
        )
    )
    ```

    Here the rate will become 10Hz again every 1 second of simulation. If the period is smaller than the schedule, the remaining rates will not be set.

    You can use the `reset()` method to manually reinitialize the schedule, times becoming relative to that call:

    ```python
    net.simulate(1200.) # Should switch to 100 Hz due to the period of 1000.
    inp.reset()
    net.simulate(1000.) # Starts at 10 Hz again.
    ```

    Note that the rates are reset to the value they had before compile().

    The rates are here common to all neurons of the population. If you want each neuron to have a different rate, `rates` must have additional dimensions corresponding to the geometry of the population. The first dimension still corresponds to the schedule.

    ```python
    inp = net.create(
        TimedPoissonPopulation(
            geometry = 100,
            rates = [ 
                [10. + 0.05*i for i in range(100)], # First 100 ms
                [20. + 0.05*i for i in range(100)], # After 100 ms
            ],
            schedule = [0., 100.],
            period = 1000.,
        )
    )
    ```

    :param rates: array of firing rates (list of floats or lists of numpy arrays). The first axis corresponds to the times where the firing rate should change and have the same length as `schedule`, if used. The other dimensions must match the geometry of the population.
    :param schedule: list of times (in ms) where the firing rate should change.
    :param period: time when the timed array will be reset and start again, allowing cycling over the schedule. Default: no cycling (-1).
    :param name: optional name for the population.
    """
    def __init__(self, geometry, rates, schedule, period= -1., name=None, copied=False, net_id=0):
        
        neuron = Neuron(
            parameters = """
            proba = 1.0
            """,
            equations = """
            p = Uniform(0.0, 1.0) * 1000.0 / dt
            """,
            spike = """
            p < proba
            """,
            name="TimedPoisson",
            description="Spiking neuron following a Poisson distribution."
        )

        SpecificPopulation.__init__(self, geometry=geometry, neuron=neuron, name=name, copied=copied, net_id=net_id)

        # Check arguments
        try:
            rates = list(rates)
        except:
            Messages._error("TimedPoissonPopulation: the rates argument must be a list of lists.")

        schedule = list(schedule)

        nb_schedules = len(rates)
        if nb_schedules != len(schedule):
            Messages._error("TimedPoissonPopulation: the first axis of the rates argument must be the same length as schedule.")


        if isinstance(rates[0], (float, int, )) : # One rate for the whole population
            rates = [np.full(self.size, rates[i]) for i in range(nb_schedules)]

        # Initial values
        self.init['schedule'] = schedule
        self.init['rates'] = rates
        self.init['period'] = period

    def _copy(self, net_id=None):
        "Returns a copy of the population when creating networks."
        return TimedPoissonPopulation(self.geometry, self.init['rates'] , self.init['schedule'], self.init['period'], self.name, copied=True, net_id=self.net_id if net_id is None else net_id)

    def _generate_st(self):
        """
        adjust code templates for the specific population for single thread.
        """
        self._specific_template['declare_additional'] = """
    // Custom local parameters of a TimedPoissonPopulation
    std::vector< int > _schedule; // List of times where new inputs should be set
    std::vector< std::vector< %(float_prec)s > > _buffer; // buffer holding the data
    int _period; // Period of cycling
    long int _t; // Internal time
    int _block; // Internal block when inputs are set not at each step
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}
        
        self._specific_template['access_additional'] = """
    // Custom local parameters of a TimedPoissonPopulation
    void set_schedule(std::vector<int> schedule) { _schedule = schedule; }
    std::vector<int> get_schedule() { return _schedule; }
    void set_rates(std::vector< std::vector< %(float_prec)s > > buffer) { _buffer = buffer; r = _buffer[0]; }
    std::vector< std::vector< %(float_prec)s > > get_rates() { return _buffer; }
    void set_period(int period) { _period = period; }
    int get_period() { return _period; }
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}
        
        self._specific_template['init_additional'] = """
        // Initialize counters
        _t = 0;
        _block = 0;
        _period = -1;
"""

        self._specific_template['reset_additional'] ="""
        _t = 0;
        _block = 0;

        r.clear();
        r = std::vector<%(float_prec)s>(size, 0.0);
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}

        self._specific_template['update_variables'] = """
        if(_active){
            //std::cout << _t << " " << _block<< " " << _schedule[_block] << std::endl;

            // Check if it is time to set the input
            if(_t == _schedule[_block]){
                // Set the data
                proba = _buffer[_block];
                // Move to the next block
                _block++;
                // If was the last block, go back to the first block
                if (_block == _schedule.size()){
                    _block = 0;
                }
            }

            // If the timedarray is periodic, check if we arrive at that point
            if(_period > -1 && (_t == _period-1)){
                // Reset the counters
                _block=0;
                _t = -1;
            }

            // Always increment the internal time
            _t++;
        }

        if( _active ) {
            spiked.clear();

            // Updating local variables
            %(float_prec)s step = 1000.0/dt;

            #pragma omp simd
            for(int i = 0; i < size; i++){

                // p = Uniform(0.0, 1.0) * 1000.0 / dt
                p[i] = step*rand_0[i];


            }
        } // active
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}

        self._specific_template['size_in_bytes'] = """
        // schedule
        size_in_bytes += _schedule.capacity() * sizeof(int);

        // buffer
        size_in_bytes += _buffer.capacity() * sizeof(std::vector<%(float_prec)s>);
        for( auto it = _buffer.begin(); it != _buffer.end(); it++ )
            size_in_bytes += it->capacity() * sizeof(%(float_prec)s);
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}
        

        self._specific_template['wrapper'] = f"""
    // TimedPoissonPopulation
    nanobind::class_<PopStruct{self.id}>(m, "pop{self.id}_wrapper")
        // Constructor
        .def(nanobind::init<int, int>())

        // Common attributes
        .def_rw("size", &PopStruct{self.id}::size)
        .def_rw("max_delay", &PopStruct{self.id}::max_delay)

        // Attributes
		.def_rw("r", &PopStruct{self.id}::r)
		.def_rw("p", &PopStruct{self.id}::p)
		.def_rw("proba", &PopStruct{self.id}::proba)

        // Access methods
        .def("set_schedule", &PopStruct{self.id}::set_schedule)
        .def("get_schedule", &PopStruct{self.id}::get_schedule)

        .def("set_rates", &PopStruct{self.id}::set_rates)
        .def("get_rates", &PopStruct{self.id}::get_rates)

        .def("set_period", &PopStruct{self.id}::set_period)
        .def("get_period", &PopStruct{self.id}::get_period)

        // Other methods
		.def("compute_firing_rate", &PopStruct{self.id}::compute_firing_rate)

        .def("activate", &PopStruct{self.id}::set_active)
        .def("reset", &PopStruct{self.id}::reset)
        .def("clear", &PopStruct{self.id}::clear);
""" 

    def _generate_omp(self):
        """
        adjust code templates for the specific population for openMP.
        """
        self._specific_template['declare_additional'] = """
    // Custom local parameters of a TimedPoissonPopulation
    std::vector< int > _schedule; // List of times where new inputs should be set
    std::vector< std::vector< %(float_prec)s > > _buffer; // buffer holding the data
    int _period; // Period of cycling
    long int _t; // Internal time
    int _block; // Internal block when inputs are set not at each step
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}
        
        self._specific_template['access_additional'] = """
    // Custom local parameters of a TimedPoissonPopulation
    void set_schedule(std::vector<int> schedule) { _schedule = schedule; }
    std::vector<int> get_schedule() { return _schedule; }
    void set_rates(std::vector< std::vector< %(float_prec)s > > buffer) { _buffer = buffer; r = _buffer[0]; }
    std::vector< std::vector< %(float_prec)s > > get_rates() { return _buffer; }
    void set_period(int period) { _period = period; }
    int get_period() { return _period; }
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}
        
        self._specific_template['init_additional'] = """
        // Initialize counters
        _t = 0;
        _block = 0;
        _period = -1;
"""

        self._specific_template['reset_additional'] ="""
        _t = 0;
        _block = 0;

        r.clear();
        r = std::vector<%(float_prec)s>(size, 0.0);
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}

        self._specific_template['update_variables'] = """
        if(_active){
            #pragma omp single
            {
                //std::cout << _t << " " << _block<< " " << _schedule[_block] << std::endl;

                // Check if it is time to set the input
                if(_t == _schedule[_block]){
                    // Set the data
                    proba = _buffer[_block];
                    // Move to the next block
                    _block++;
                    // If was the last block, go back to the first block
                    if (_block == _schedule.size()){
                        _block = 0;
                    }
                }

                // If the timedarray is periodic, check if we arrive at that point
                if(_period > -1 && (_t == _period-1)){
                    // Reset the counters
                    _block=0;
                    _t = -1;
                }

                // Always increment the internal time
                _t++;
            }
        }

        if( _active ) {
            spiked.clear();

            // Updating local variables
            %(float_prec)s step = 1000.0/dt;

            #pragma omp for simd
            for(int i = 0; i < size; i++){

                // p = Uniform(0.0, 1.0) * 1000.0 / dt
                p[i] = step*rand_0[i];


            }
        } // active
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}

        self._specific_template['size_in_bytes'] = """
        // schedule
        size_in_bytes += _schedule.capacity() * sizeof(int);

        // buffer
        size_in_bytes += _buffer.capacity() * sizeof(std::vector<%(float_prec)s>);
        for( auto it = _buffer.begin(); it != _buffer.end(); it++ )
            size_in_bytes += it->capacity() * sizeof(%(float_prec)s);
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}
        

        self._specific_template['wrapper'] = f"""
    // TimedPoissonPopulation
    nanobind::class_<PopStruct{self.id}>(m, "pop{self.id}_wrapper")
        // Constructor
        .def(nanobind::init<int, int>())

        // Common attributes
        .def_rw("size", &PopStruct{self.id}::size)
        .def_rw("max_delay", &PopStruct{self.id}::max_delay)

        // Attributes
		.def_rw("r", &PopStruct{self.id}::r)
		.def_rw("p", &PopStruct{self.id}::p)
		.def_rw("proba", &PopStruct{self.id}::proba)

        // Access methods
        .def("set_schedule", &PopStruct{self.id}::set_schedule)
        .def("get_schedule", &PopStruct{self.id}::get_schedule)

        .def("set_rates", &PopStruct{self.id}::set_rates)
        .def("get_rates", &PopStruct{self.id}::get_rates)

        .def("set_period", &PopStruct{self.id}::set_period)
        .def("get_period", &PopStruct{self.id}::get_period)

        // Other methods
		.def("compute_firing_rate", &PopStruct{self.id}::compute_firing_rate)

        .def("activate", &PopStruct{self.id}::set_active)
        .def("reset", &PopStruct{self.id}::reset)
        .def("clear", &PopStruct{self.id}::clear);
"""

    def _generate_cuda(self):
        """
        Code generation if the CUDA paradigm is set.
        """
        # I suppress the code generation for allocating the variable r on gpu, as
        # well as memory transfer codes. This is only possible as no other variables
        # allowed in TimedArray.
        self._specific_template['init_parameters_variables'] = """
        // Random numbers
        cudaMalloc((void**)&gpu_rand_0, size * sizeof(curandState));
        init_curand_states( size, gpu_rand_0, global_seed );
"""
        self._specific_template['host_device_transfer'] = ""
        self._specific_template['device_host_transfer'] = ""

        #
        # Code for handling the buffer and schedule parameters
        self._specific_template['declare_additional'] = """
    // Custom local parameter timed array
    std::vector< int > _schedule;
    std::vector< %(float_prec)s* > gpu_buffer;
    int _period; // Period of cycling
    long int _t; // Internal time
    int _block; // Internal block when inputs are set not at each step
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}
        self._specific_template['access_additional'] = """
    // Custom local parameter timed array
    void set_schedule(std::vector<int> schedule) { _schedule = schedule; }
    std::vector<int> get_schedule() { return _schedule; }
    void set_rates(std::vector< std::vector< %(float_prec)s > > buffer) {
        if ( gpu_buffer.empty() ) {
            gpu_buffer = std::vector< %(float_prec)s* >(buffer.size(), nullptr);
            // allocate gpu arrays
            for(int i = 0; i < buffer.size(); i++) {
                cudaMalloc((void**)&gpu_buffer[i], buffer[i].size()*sizeof(%(float_prec)s));
            }
        }

        auto host_it = buffer.begin();
        auto dev_it = gpu_buffer.begin();
        for( ; host_it < buffer.end(); host_it++, dev_it++ ) {
            cudaMemcpy( *dev_it, host_it->data(), host_it->size()*sizeof(%(float_prec)s), cudaMemcpyHostToDevice);
        }

        gpu_proba = gpu_buffer[0];
    }

    std::vector< std::vector< %(float_prec)s > > get_rates() {
        std::vector< std::vector< %(float_prec)s > > buffer = std::vector< std::vector< %(float_prec)s > >( gpu_buffer.size(), std::vector<%(float_prec)s>(size,0.0) );

        auto host_it = buffer.begin();
        auto dev_it = gpu_buffer.begin();
        for( ; host_it < buffer.end(); host_it++, dev_it++ ) {
            cudaMemcpy( host_it->data(), *dev_it, size*sizeof(%(float_prec)s), cudaMemcpyDeviceToHost );
        }

        return buffer;
    }
    void set_period(int period) { _period = period; }
    int get_period() { return _period; }
""" % {'float_prec': ConfigManager().get('precision', self.net_id)}
        
        self._specific_template['init_additional'] = """
        // counters
        _t = 0;
        _block = 0;
        _period = -1;
"""

        self._specific_template['reset_additional'] = """
        // counters
        _t = 0;
        _block = 0;
        gpu_proba = gpu_buffer[0];
"""

        self._specific_template['update_variables'] = """
        if(_active) {
            // std::cout << _t << " " << _block<< " " << _schedule[_block] << std::endl;
            // Check if it is time to set the input
            if(_t == _schedule[_block]){
                // Set the data
                gpu_proba = gpu_buffer[_block];
                // Move to the next block
                _block++;
                // If was the last block, go back to the first block
                if ( _block == _schedule.size() ) {
                    _block = 0;
                }
            }

            // If the timedarray is periodic, check if we arrive at that point
            if( (_period > -1) && (_t == _period-1) ) {
                // Reset the counters
                _block=0;
                _t = -1;
            }

            // Always increment the internal time
            _t++;
        }
"""

        self._specific_template['update_variable_body'] = """
__global__ void cuPop%(id)s_local_step( const long int t, const double dt, curandState* rand_0, double* proba, unsigned int* num_events, int* spiked, long int* last_spike )
{
    int i = threadIdx.x + blockDim.x * blockIdx.x;
    %(float_prec)s step = 1000.0/dt;

    while ( i < %(size)s )
    {
        // p = Uniform(0.0, 1.0) * 1000.0 / dt
        %(float_prec)s p = curand_uniform_double( &rand_0[i] ) * step;

        if (p < proba[i]) {
            int pos = atomicAdd ( num_events, 1);
            spiked[pos] = i;
            last_spike[i] = t;
        }

        i += blockDim.x;
    }

    __syncthreads();
}
""" % {
    'id': self.id,
    'size': self.size,
    'float_prec': ConfigManager().get('precision', self.net_id)
}

        self._specific_template['update_variable_header'] = "__global__ void cuPop%(id)s_local_step( const long int t, const double dt, curandState* rand_0, double* proba, unsigned int* num_events, int* spiked, long int* last_spike );" % {'id': self.id}
        # Please notice, that the GPU kernels can be launched only with one block. Otherwise, the
        # atomicAdd which is called inside the kernel is not working correct (HD: April 1st, 2021)
        self._specific_template['update_variable_call'] = """
    // host side update of neurons
    pop%(id)s->update();

    // Reset old events
    clear_num_events<<< 1, 1, 0, pop%(id)s->stream >>>(pop%(id)s->gpu_spike_count);
#ifdef _DEBUG
    cudaError_t err_clear_num_events_%(id)s = cudaGetLastError();
    if(err_clear_num_events_%(id)s != cudaSuccess)
        std::cout << "pop%(id)s_spike_gather: " << cudaGetErrorString(err_clear_num_events_%(id)s) << std::endl;
#endif

    // Compute current events
    cuPop%(id)s_local_step<<< 1, pop%(id)s->_threads_per_block, 0, pop%(id)s->stream >>>(
        t, dt,
        pop%(id)s->gpu_rand_0,
        pop%(id)s->gpu_proba,
        pop%(id)s->gpu_spike_count,
        pop%(id)s->gpu_spiked,
        pop%(id)s->gpu_last_spike
    );
#ifdef _DEBUG
    cudaError_t err_pop_spike_gather_%(id)s = cudaGetLastError();
    if(err_pop_spike_gather_%(id)s != cudaSuccess)
        std::cout << "pop%(id)s_spike_gather: " << cudaGetErrorString(err_pop_spike_gather_%(id)s) << std::endl;
#endif

    // transfer back the spike counter (needed by record)
    cudaMemcpyAsync( &pop%(id)s->spike_count, pop%(id)s->gpu_spike_count, sizeof(unsigned int), cudaMemcpyDeviceToHost, pop%(id)s->stream );
#ifdef _DEBUG
    cudaError_t err = cudaGetLastError();
    if ( err != cudaSuccess )
        std::cout << "record_spike_count: " << cudaGetErrorString(err) << std::endl;
#endif

    // transfer back the spiked array (needed by record)
    cudaMemcpyAsync( pop%(id)s->spiked.data(), pop%(id)s->gpu_spiked, pop%(id)s->spike_count*sizeof(int), cudaMemcpyDeviceToHost, pop%(id)s->stream );
#ifdef _DEBUG
    err = cudaGetLastError();
    if ( err != cudaSuccess )
        std::cout << "record_spike: " << cudaGetErrorString(err) << std::endl;
#endif

""" % {'id': self.id}

        self._specific_template['size_in_bytes'] = "//TODO: "



        self._specific_template['wrapper'] = f"""
    // TimedPoissonPopulation
    nanobind::class_<PopStruct{self.id}>(m, "pop{self.id}_wrapper")
        // Constructor
        .def(nanobind::init<int, int>())

        // Common attributes
        .def_rw("size", &PopStruct{self.id}::size)
        .def_rw("max_delay", &PopStruct{self.id}::max_delay)

        // Attributes
		.def_rw("r", &PopStruct{self.id}::r)
		.def_rw("p", &PopStruct{self.id}::p)
		.def_rw("proba", &PopStruct{self.id}::proba)

        // Access methods
        .def("set_schedule", &PopStruct{self.id}::set_schedule)
        .def("get_schedule", &PopStruct{self.id}::get_schedule)

        .def("set_rates", &PopStruct{self.id}::set_rates)
        .def("get_rates", &PopStruct{self.id}::get_rates)

        .def("set_period", &PopStruct{self.id}::set_period)
        .def("get_period", &PopStruct{self.id}::get_period)

        // Other methods
		.def("compute_firing_rate", &PopStruct{self.id}::compute_firing_rate)

        .def("activate", &PopStruct{self.id}::set_active)
        .def("reset", &PopStruct{self.id}::reset)
        .def("clear", &PopStruct{self.id}::clear);
""" 

    def _instantiate(self, module):
        # Create the Cython instance
        self.cyInstance = getattr(module, self.class_name+'_wrapper')(self.size, self.max_delay)

    def __setattr__(self, name, value):
        if name == 'schedule':
            if self.initialized:
                self.cyInstance.set_schedule( [int(val / ConfigManager().get('dt', self.net_id))  for val in value])
            else:
                self.init['schedule'] = value
        elif name == 'rates':
            if self.initialized:

                value = list(value)
                
                if isinstance(value[0], (float, int)): # same value for each neuron, create a list of list
                    value = [ [float(value[i]) for _ in range(self.size)] for i in range(len(value)) ]
                else:
                    value = [list(np.array(value[i]).flatten()) for i in range(len(value))]
                
                self.cyInstance.set_rates(value)

            else:
                self.init['rates'] = value
        elif name == "period":
            if self.initialized:
                self.cyInstance.set_period(int(value /ConfigManager().get('dt', self.net_id)))
            else:
                self.init['period'] = value
        else:
            Population.__setattr__(self, name, value)

    def __getattr__(self, name):
        if name == 'schedule':
            if self.initialized:
                return [float(ConfigManager().get('dt', self.net_id) * val) for val in self.cyInstance.get_schedule()]
            else:
                return self.init['schedule']
        elif name == 'rates':
            if self.initialized:
                if len(self.geometry) > 1:
                    # unflatten the data
                    flat_values = self.cyInstance.get_rates()
                    values = []
                    for x in range(len(self.schedule)):
                        values.append(np.reshape(flat_values[x], self.geometry))
                    return values
                else:
                    return self.cyInstance.get_rates()
            else:
                return self.init['rates']
        elif name == 'period':
            if self.initialized:
                return self.cyInstance.get_period() * ConfigManager().get('dt', self.net_id)
            else:
                return self.init['period']
        else:
            return Population.__getattribute__(self, name)


"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

attribute_decl = {
    'local':
"""
    // Local %(attr_type)s %(name)s
    std::vector< %(type)s > %(name)s;
""",
    'semiglobal':
"""
    // Semiglobal %(attr_type)s %(name)s
    std::vector< %(type)s >  %(name)s ;
""",
    'global':
"""
    // Global %(attr_type)s %(name)s
    %(type)s  %(name)s ;
"""
}

attribute_cpp_init = {
    'local':
"""
        // Local %(attr_type)s %(name)s
        %(name)s = init_matrix_variable<%(type)s>(static_cast<%(type)s>(%(init)s));
""",
    'semiglobal':
"""
        // Semiglobal %(attr_type)s %(name)s
        %(name)s = init_vector_variable<%(type)s>(static_cast<%(type)s>(%(init)s));
""",
    'global':
"""
        // Global %(attr_type)s %(name)s
        %(name)s = %(init)s;
"""
}

attribute_cpp_size = {
    'local': """
        // Local %(attr_type)s %(name)s
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();
""",
    'semiglobal': """
        // Semiglobal %(attr_type)s %(name)s
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();
""",
    'global': """
        // Global %(attr_type)s %(name)s
        size_in_bytes += sizeof(%(ctype)s);
"""
}

attribute_cpp_delete = {
    'local': """
        // %(name)s
        %(name)s.clear();
        %(name)s.shrink_to_fit();
""",
    'semiglobal': """
        // %(name)s
        %(name)s.clear();
        %(name)s.shrink_to_fit();
""",
    'global': ""
}

delay = {
    # A single value for all synapses
    'uniform': {
        'declare': """
    // Uniform delay
    int delay;

    int get_delay() { return delay; }
    int get_dendrite_delay(int idx) { return delay; }
    void set_delay(int delay) { this->delay = delay; }
""",
        'init': """
    delay = delays[0][0];
"""
    },
    # An individual value for each synapse
    'nonuniform_rate_coded': None,
    # An individual value for each synapse and a
    # buffer for spike events    
    'nonuniform_spiking': None
}

event_driven = {
    'declare': """
    std::vector<long> _last_event;
""",
    'cpp_init': """
        // Event-driven
        _last_event = init_matrix_variable<long>(-10000);
""",
    'pyx_struct': """
        vector[vector[long]] _last_event
""",
}

spiking_summation_fixed_delay = """// Event-based summation
if (_transmission && %(post_prefix)s_active) {
    const int post_slice_beg_ = post_slices_[tid];
    const int post_slice_end_ = post_slices_[tid+1];
#ifdef _DEBUG_OMP_DETAIL
    #pragma omp critical
    {
        std::cout << "thread " << tid << ": " << post_slice_beg_ << " - " << post_slice_end_ << std::endl;
    }
#endif

    // Iterate over all spiking neurons
    for (auto it = %(pre_prefix)sspiked.cbegin(); it != %(pre_prefix)sspiked.cend(); it++) {
        %(idx_type)s rk_pre = (*it);
        %(size_type)s beg = rk_pre * this->num_rows_;
        %(size_type)s end = (rk_pre+1) * this->num_rows_;

        // Iterate over columns
        for (%(idx_type)s rk_post = post_slice_beg_; rk_post < post_slice_end_; rk_post++) {
            %(size_type)s j = beg + rk_post;
            
            %(g_target)s

            if (mask_[j]) {
                %(event_driven)s
                %(pre_event)s
            }
        }
    }
} // active
"""

spiking_summation_fixed_delay_only_psp_inner_loop = """// Event-based summation
if (_transmission && %(post_prefix)s_active) {
    const int post_slice_beg_ = post_slices_[tid];
    const int post_slice_end_ = post_slices_[tid+1];
#ifdef _DEBUG_OMP_DETAIL
    #pragma omp critical
    {
        std::cout << "thread " << tid << ": " << post_slice_beg_ << " - " << post_slice_end_ << std::endl;
    }
#endif

    // Iterate over all spiking neurons
    for (auto it = %(pre_prefix)sspiked.cbegin(); it != %(pre_prefix)sspiked.cend(); it++) {
        %(idx_type)s rk_pre = (*it);
        %(size_type)s beg = rk_pre * this->num_rows_;
        %(size_type)s end = (rk_pre+1) * this->num_rows_;

        // Iterate over columns
        for (%(idx_type)s rk_post = post_slice_beg_; rk_post < post_slice_end_; rk_post++) {
            %(size_type)s j = beg + rk_post;
            %(g_target)s
        }
    }
} // active
"""

dense_update_variables = {
    'local': """
// Check periodicity
if(_transmission && _update && %(post_prefix)s_active && ( (t - _update_offset)%%_update_period == 0L)){
    // Global variables
    %(global)s

    // Local variables
    #pragma omp for
    for(%(idx_type)s i = 0; i < %(post_prefix)ssize; i++){
        rk_post = i; // dense: ranks are indices

        // Semi-global variables
    %(semiglobal)s

        // Local variables are updated to boolean flag
        %(size_type)s j = i*%(pre_prefix)ssize;
        for(rk_pre = 0; rk_pre < %(pre_prefix)ssize; rk_pre++, j++) {
            if(mask_[j]) {
%(local)s
            }
        }
    }
}
""",
    'global': """
// Check periodicity
if(_transmission && _update && %(post_prefix)s_active && ( (t - _update_offset)%%_update_period == 0L)){
    // Global variables
    %(global)s

    // Semi-global variables
    #pragma omp for
    for(int i = 0; i < %(post_prefix)ssize; i++){
        rk_post = i;
    %(semiglobal)s
    }
}
"""
}

spiking_post_event = """
if (_transmission && %(post_prefix)s_active) {

    %(idx_type)s rows = pop%(id_pre)s->size;
    %(idx_type)s columns = pop%(id_post)s->size;

    for (%(idx_type)s _idx_i = tid; _idx_i < %(post_prefix)sspiked.size(); _idx_i+= nt) {
        %(idx_type)s post_rank = %(post_prefix)sspiked[_idx_i];
        %(idx_type)s rk_pre = 0;

        for (%(size_type)s j = post_rank; j < this->num_rows_ * this->num_columns_; j += this->num_rows_, rk_pre++) {
            if(mask_[j]) {
%(event_driven)s
%(post_event)s
            }
        }
    }
}
"""

conn_templates = {
    # accessors
    'attribute_decl': attribute_decl,
    'attribute_cpp_init': attribute_cpp_init,
    'attribute_cpp_size': attribute_cpp_size,
    'attribute_cpp_delete': attribute_cpp_delete,
    'delay': delay,
    'event_driven': event_driven,

    #operations
    'rate_coded_sum': "",
    'vectorized_default_psp': {},
    'spiking_sum_fixed_delay': {
        # HD: need to distinguish ????
        'inner_loop': spiking_summation_fixed_delay,
        'outer_loop': spiking_summation_fixed_delay
    },
    'spiking_sum_fixed_delay_only_psp': {
        'inner_loop': spiking_summation_fixed_delay_only_psp_inner_loop,
        'outer_loop': None,
    },
    'update_variables': dense_update_variables,
    'post_event': spiking_post_event
}

conn_ids = {
    'local_index': '[j]',
    'semiglobal_index': '[i]',
    'global_index': '',
    'post_index': '[rk_post]',
    'pre_index': '[rk_pre]',
    'delay_u' : '[delay-1]' # uniform delay
}



#!/usr/bin/env python
import sys
import os, os.path
import shutil
import json
import subprocess
import numpy

from setuptools import setup, Extension
from setuptools.command.build import build
from Cython.Build import cythonize

#########################################################
# ANNarchy global configurations (during installation)
#########################################################
def create_config():
    """ Creates config file and check for typical configurations. """

    def cuda_config():
        cuda_path = "/usr/local/cuda"
        if os.path.exists('/usr/local/cuda-7.0'):
            cuda_path = "/usr/local/cuda-7.0"
        if os.path.exists('/usr/local/cuda-8.0'):
            cuda_path = "/usr/local/cuda-8.0"
        if os.path.exists('/usr/local/cuda-9.0'):
            cuda_path = "/usr/local/cuda-9.0"
        if os.path.exists('/usr/local/cuda-10.0'):
            cuda_path = "/usr/local/cuda-10.0"
        return {
            'compiler': "nvcc",
            'flags': "",
            'device': 0,
            'path': cuda_path}

    # Generate a default setting files
    settings = {}

    # OpenMP settings
    if sys.platform.startswith('linux'): # Linux systems
        settings['openmp'] = {
            'compiler': "g++",
            'flags': "-O3 -march=native",
        }
    elif sys.platform == "darwin":   # mac os
        settings['openmp'] = {
            'compiler': "clang++",
            'flags': "-O3",
        }

    # CUDA settings (optional)
    settings['cuda'] = cuda_config()

    # If the config file does not exist, create it
    if not os.path.exists(os.path.expanduser('~/.config/ANNarchy/annarchy.json')):
        print('Creating the configuration file in ~/.config/ANNarchy/annarchy.json')
        if not os.path.exists(os.path.expanduser('~/.config/ANNarchy')):
            os.makedirs(os.path.expanduser('~/.config/ANNarchy'))

        # Write the settings
        with open(os.path.expanduser("~/.config/ANNarchy/annarchy.json"), 'w') as f:
            json.dump(settings, f, indent=4)

        return settings

    else: # The config file exists, make sure it has all the required fields
        update_required = False
        print_hint_msg = True
        with open(os.path.expanduser("~/.config/ANNarchy/annarchy.json"), 'r') as f:
            # Load the existing settings
            local_settings = json.load(f)
            # Check the first level headers exist
            for paradigm in settings.keys():
                if not paradigm in local_settings.keys():
                    local_settings[paradigm] = settings[paradigm]
                    update_required = True
                    break
                # Iterate over all obligatory fields and update then if required
                for key in settings[paradigm].keys():
                    if not key in local_settings[paradigm].keys():
                        local_settings[paradigm][key] = settings[paradigm][key]
                        update_required = True
                    else:
                        # Check if the set values diverge
                        if local_settings[paradigm][key] != settings[paradigm][key]:
                            if print_hint_msg:
                                print("HINT - we detected diverging settings:")
                                print_hint_msg = False
                            print("   * field ["+paradigm+","+key+"]:")
                            print("       - your annarchy.json: ", local_settings[paradigm][key] )
                            print("       - ANNarchy default: ", settings[paradigm][key] )

        if update_required:
            print('Updating the configuration file in ~/.config/ANNarchy/annarchy.json')
            with open(os.path.expanduser("~/.config/ANNarchy/annarchy.json"), 'w') as f:
                json.dump(local_settings, f, indent=4)

        return local_settings

def python_environment():
    """
    Python environment configuration. Detects among others the python version, library path and cython version.
    """
    # Python version
    py_version = "%(major)s.%(minor)s" % {'major': sys.version_info[0],
                                          'minor': sys.version_info[1]}
    py_major = str(sys.version_info[0])

    # Python includes and libs
    py_prefix = sys.base_prefix

    # Search for pythonx.y-config
    cmd = "%(py_prefix)s/bin/python%(py_version)s-config --includes > /dev/null 2> /dev/null"
    with subprocess.Popen(cmd % {'py_version': py_version, 'py_prefix': py_prefix}, shell=True) as test:
        if test.wait() != 0:
            print("Can not find python-config in the same directory as python, trying with the default path...")
            python_config_path = "python%(py_version)s-config"% {'py_version': py_version}
        else:
            python_config_path = "%(py_prefix)s/bin/python%(py_version)s-config" % {'py_version': py_version, 'py_prefix': py_prefix}

    python_include = "`%(pythonconfigpath)s --includes`" % {'pythonconfigpath': python_config_path}
    python_libpath = "-L%(py_prefix)s/lib" % {'py_prefix': py_prefix}

    # Check cython version
    with subprocess.Popen(py_prefix + "/bin/cython%(major)s -V > /dev/null 2> /dev/null" % {'major': py_major}, shell=True) as test:
        if test.wait() != 0:
            cython = py_prefix + "/bin/cython"
        else:
            cython = py_prefix + "/bin/cython" + py_major

    # If not in the same folder as python, use the default
    with subprocess.Popen("%(cython)s -V > /dev/null 2> /dev/null" % {'cython': cython}, shell=True) as test:
        if test.wait() != 0:
            cython = shutil.which("cython"+py_major)
            if cython is None:
                cython = shutil.which("cython")
                if cython is None:
                    print("ERROR: Unable to find the 'cython' executable, fix your $PATH if already installed." )
                    sys.exit(1)

    return py_version, py_major, python_include, python_libpath, cython

# Extra compile args
extra_compile_args = ["-O3"]
extra_link_args = []
if sys.platform.startswith('darwin'):
    extra_compile_args.append("-stdlib=libc++")
    extra_link_args = ["-stdlib=libc++"]

################################################
# Perform the installation
################################################
package_data = [
                'cython_ext/*.pxd',
                'cython_ext/*.pyx',
                'cython_ext/CSRMatrix.hpp',
                'include/*.hpp',
                'include/*.cuh',
                'thirdparty/*.hpp'
                ]

extensions = [
    Extension("ANNarchy.cython_ext.Connector",
            ["ANNarchy/cython_ext/Connector.pyx"],
            include_dirs=[numpy.get_include()],
            extra_compile_args=extra_compile_args,
            extra_link_args=extra_link_args,
            language="c++"),
    Extension("ANNarchy.cython_ext.Coordinates",
            ["ANNarchy/cython_ext/Coordinates.pyx"],
            include_dirs=[numpy.get_include()],
            extra_compile_args=extra_compile_args,
            extra_link_args=extra_link_args,
            language="c++"),
    Extension("ANNarchy.cython_ext.Transformations",
            ["ANNarchy/cython_ext/Transformations.pyx"],
            include_dirs=[numpy.get_include()],
            extra_compile_args=extra_compile_args,
            extra_link_args=extra_link_args,
            language="c++"),
]

class CustomizedBuild(build):
    """
    Customization of the build process.
    """

    def run(self):
        """
        Extend the default build step
        """
        # Own stuff
        print("Installing ANNarchy ...")
        py_version, _, python_include, python_libpath, cython_major = python_environment()
        print("\tPython", py_version, "(", sys.executable, ')')
        print("\tIncludes:", python_include)
        print("\tLibraries:", python_libpath)
        print("\tCython:", cython_major)
        print("\tNumpy:", numpy.get_include())

        print("Check annarchy.json")
        create_config()

        print("Building Cython Extensions ...")
        # perform default behavior
        build.run(self)     # NEVER call super, it breaks everything!

# PyExtension stuff remains here while the rest of metadata is contained in
# pyproject.toml ...
setup(
    ext_modules = cythonize(extensions, language_level=int(sys.version_info[0])),
    package_data={'ANNarchy': package_data},
    cmdclass={"build": CustomizedBuild}
)



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.intern.NetworkManager import NetworkManager
from ANNarchy.intern import Messages

class GlobalObjectManager:
    """
    Helper class to ensure, that the neurons, synapses and functions are globally accessible.
    """
    # singleton instance
    _instance = None
    # storage
    _objects = None

    def __init__(self):
        """
        Constructor.
        """
        pass

    def __new__(cls):
        """
        First call construction of the GlobalObjectManager. No additional arguments are required.
        """
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._create_initial_state(cls)
        
        return cls._instance

    def _create_initial_state(self):
        """
        Initialize the container for globally available objects.

        Called either from __new__ or clear(). The first
        slot is reserved for the magic network.
        """
        self._objects = {
            'neurons': [],
            'synapses': [],
            'functions': []
        }

    def clear(self, functions:bool=True, neurons:bool=True, synapses:bool=True):
        """
        Remove all instantiated objects.
        """
        if functions:
            self._objects['functions'] = []

        if neurons:
            for obj in self._objects['neurons']:
                del obj
            self._objects['neurons'] = []

        if synapses:
            for obj in self._objects['synapses']:
                del obj
            self._objects['synapses'] = []

    ################################
    ## Neuron types
    ################################
    def add_neuron_type(self, neuron):
        self._objects['neurons'].append(neuron)

    def get_neuron_types(self):
        return self._objects['neurons']

    def num_neuron_types(self):
        return len(self._objects['neurons'])

    ################################
    ## Synapse types
    ################################
    def add_synapse_type(self, synapse):
        self._objects['synapses'].append(synapse)

    def get_synapse_types(self):
        return self._objects['synapses']

    def num_synapse_types(self):
        return len(self._objects['synapses'])


    ################################
    ## Functions
    ################################
    def add_function(self, function):
        name = function.split('(')[0]
        if self.get_function(name) is not None:
            Messages._error(f"add_function(): the global function {name} already exists at the global level.")
        self._objects['functions'].append( (name, function) )

    def functions(self, name:str, network=None):

        net_id = 0 if network is None else network.id
        try:
            func = getattr(NetworkManager().get_network(net_id=net_id).instance, 'func_' + name)
        except:
            Messages._error('call to', name, ': the function is not compiled yet.')

        return func

    def get_function(self, name):
        """
        Returns the function string with the provided name.
        """
        for n, function in self._objects['functions']:
            if n == name:
                return name
        return None

    def get_functions(self):
        """
        Returns all functions
        """
        return self._objects['functions']

    def number_functions(self):
        return len(self._objects['functions'])


    ################################
    ## Constants
    ################################

    def list_constants(self):
        """
        Only for the parser, constants are stored in networks
        """
        names = []
        for network in NetworkManager().get_networks():
            if network is None:
                # The network was either already deleted or not completely
                # initialized ...
                continue

            constants = network.get_constants()
            for c in constants:
                names.append(c.name)

        return list(set(names))



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.intern.ConfigManagement import ConfigManager
from ANNarchy.intern import Messages
from ANNarchy.core.Random import RandomDistribution

from typing import Iterator
import numpy as np

class Dendrite :
    """
    Sub-group of a `Projection` for a single post-synaptic neuron.

    It can not be created directly, only through a call to ``Projection.dendrite(rank)``:

    ```python
    dendrite = proj.dendrite(6)
    ```
    """
    def __init__(self, proj, post_rank, idx):

        self.post_rank = post_rank
        "Rank of the post-synaptic neuron."
        self.idx = idx
        self.proj = proj
        "Parent projection."
        self.pre = proj.pre

        self.target = self.proj.target

        self.attributes = self.proj.attributes
        self.parameters = self.proj.parameters
        self.variables = self.proj.variables

    @property
    def size(self) -> int:
        """
        Number of synapses reaching the post-synaptic neuron.
        """
        if self.proj.cyInstance:
            return self.proj.cyInstance.dendrite_size(self.idx)
        return 0

    @property
    def pre_ranks(self) -> list[int]:
        """
        List of ranks of pre-synaptic neurons.
        """
        if self.proj.cyInstance:
            return self.proj.cyInstance.pre_rank(self.idx)
        return []

    def __len__(self):
        # Number of synapses.
        return self.size

    @property
    def synapses(self) -> Iterator["IndividualSynapse"]:
        """
        Iteratively returns the synapses corresponding to this dendrite.
        """
        for n in self.pre_ranks:
            yield IndividualSynapse(self, n)

    def synapse(self, pos:int|tuple[int]) -> "IndividualSynapse":
        """
        Returns the synapse coming from the corresponding presynaptic neuron.

        :param pos: can be either the rank or the coordinates of the presynaptic neuron
        """
        if isinstance(pos, int):
            rank = pos
        else:
            rank = self.proj.pre.rank_from_coordinates(pos)

        if rank in self.pre_ranks:
            return IndividualSynapse(self, rank)
        else:
            Messages._error(" The neuron of rank "+ str(rank) + " has no synapse in this dendrite.")
            return None

    # Iterators
    def __getitem__(self, *args, **kwds):
        # Returns the synapse of the given position in the presynaptic population.
        # If only one argument is given, it is a rank. If it is a tuple, it is coordinates.
        
        if len(args) == 1:
            return self.synapse(args[0])
        return self.synapse(args)

    def __iter__(self):
        # Returns iteratively each synapse in the dendrite in ascending pre-synaptic rank order.
        for n in self.pre_ranks:
            yield IndividualSynapse(self, n)

    #########################
    ### Access to attributes
    #########################
    def __getattr__(self, name):
        # Method called when accessing an attribute.
        if name == 'proj':
            return object.__getattribute__(self, name)
        elif hasattr(self, 'proj'):
            if name == 'rank': # TODO: remove 'rank' in a future version
                Messages._warning("Dendrite.rank: the attribute is deprecated, use Dendrite.pre_ranks instead.")
                return self.proj.cyInstance.pre_rank(self.idx)
            
            elif name == 'pre_rank':
                return self.proj.cyInstance.pre_rank(self.idx)
            
            elif name == 'delay':
                if self.proj.uniform_delay == -1:
                    return [d*ConfigManager().get('dt', self.proj.net_id) for d in self.proj.cyInstance.get_dendrite_delay(self.idx)]
                else:
                    return self.proj.max_delay * ConfigManager().get('dt', self.proj.net_id)
            
            elif name == "w" and self.proj._has_single_weight():
                return getattr(self.proj.cyInstance, "get_global_attribute_"+ConfigManager().get('precision', self.proj.net_id))(name)
            
            elif name in self.proj.attributes:
                # Determine C++ data type
                ctype = None
                for var in self.proj.synapse_type.description['variables']+self.proj.synapse_type.description['parameters']:
                    if var['name'] == name:
                        ctype = var['ctype']

                if name in self.proj.synapse_type.description['local']:
                    return getattr(self.proj.cyInstance, "get_local_attribute_row_"+ctype)(name, self.idx)
                elif name in self.proj.synapse_type.description['semiglobal']:
                    return getattr(self.proj.cyInstance, "get_semiglobal_attribute_"+ctype)(name, self.idx)
                else:
                    return getattr(self.proj.cyInstance, "get_global_attribute_"+ctype)(name)
            else:
                return object.__getattribute__(self, name)
        else:
            return object.__getattribute__(self, name)

    def __setattr__(self, name, value):
        # Method called when setting an attribute.
        if name == 'proj':
            object.__setattr__(self, 'proj', value)
        elif name == 'attributes':
            object.__setattr__(self, 'attributes', value)
        elif hasattr(self, 'proj'):
            if name in self.proj.attributes:
                # Determine C++ data type
                ctype = None
                for var in self.proj.synapse_type.description['variables']+self.proj.synapse_type.description['parameters']:
                    if var['name'] == name:
                        ctype = var['ctype']

                if name in self.proj.synapse_type.description['local']:
                    if isinstance(value, (np.ndarray, list)):
                        getattr(self.proj.cyInstance, "set_local_attribute_row_"+ctype)(name, self.idx, list(value))
                    else:
                        getattr(self.proj.cyInstance, "set_local_attribute_row_"+ctype)(name, self.idx, [value for _ in range(self.size)])

                elif name in self.proj.synapse_type.description['semiglobal']:
                    getattr(self.proj.cyInstance, "set_semiglobal_attribute_"+ctype)(name, self.idx, value)

                else:
                    # HD: will break the execution of the program
                    Messages._error("Projection attributes marked as *projection* should not be updated through dendrites.")
            else:
                object.__setattr__(self, name, value)
        else:
            object.__setattr__(self, name, value)

    def set(self, value:dict) -> None:
        """
        Sets the value of a parameter/variable on all synapses in the dendrite.

        ```python
        dendrite.set( { 'tau' : 20, 'w'= Uniform(0.0, 1.0) } )
        ```

        :param value: a dictionary containing the parameter/variable names as keys.
        """
        for key, value in value.items():
            # sanity check and then forward to __setattr__
            if key in self.attributes:
                setattr(self, key, value)
            else:
                Messages._error("Dendrite has no parameter/variable called", key)

    def get(self, name:str) -> float:
        """
        Returns the value of a parameter/variable.

        ```python
        dendrite.get('w')
        ```

        :param name: name of the parameter/variable.
        """
        if name == 'rank':
            Messages._warning("Dendrite.get('rank'): the attribute is deprecated, use Dendrite.pre_ranks instead.")
            return self.proj.cyInstance.pre_rank(self.idx)
        elif name == 'pre_ranks':
            return self.proj.cyInstance.pre_rank(self.idx)
        elif name in self.attributes:
            return getattr(self, name)
        else:
            Messages._error("Dendrite has no parameter/variable called", name)


    #########################
    ### Formatting
    #########################
    def receptive_field(self, variable:str='w', fill:float=0.0) -> np.array:
        """
        Returns the given variable as a receptive field.

        A numpy array of the same geometry as the pre-synaptic population is returned. 
        Non-existing synapses are replaced by zeros (or the value ``fill``).

        :param variable: name of the variable (default = 'w')
        :param fill: value to use when a synapse does not exist (default: 0.0).
        """
        values = getattr(self.proj.cyInstance, 'get_dendrite_'+variable)(self.idx)
        pre_ranks = self.proj.cyInstance.pre_rank( self.idx )

        m = fill * np.ones( self.pre.size )
        m[pre_ranks] = values

        return m.reshape(self.pre.geometry)


    #########################
    ### Structural plasticity
    #########################
    def create_synapse(self, rank:int, w:float=0.0, delay:float=0) -> None:
        """
        Creates a single synapse for this dendrite with the given pre-synaptic neuron.

        The configuration key `'structural_plasticity'` must be set to `True` before `compile()` for this method to work. 

        ```python
        net = ann.Network()
        net.config(structural_plasticity=True)
        net.compile()

        try:
            proj.dendrite(10).create_synapse(rank=20, w=0.1, delay=0.0)
        except Exception as e:
            print(e)
        ```

        If the synapse already exists, an error is thrown, so make sure to catch the exception.

        :param rank: rank of the pre-synaptic neuron
        :param w: synaptic weight.
        :param delay: synaptic delay in milliseconds that should be a multiple of *dt*.
        """
        if not ConfigManager().get('structural_plasticity', self.proj.net_id):
            Messages._error('"structural_plasticity" has not been set to True in setup(), can not add the synapse.')
            return

        if self.proj.cyInstance.synapse_exists(self.post_rank, rank):
            Messages._error(f'Dendrite.create_synapse(): The synapse of rank {rank} already exists for the dendrite {self.post_rank}.')
            return

        # If not all neurons in the post-synaptic population receive connections
        # the post-rank diverge from the LIL index
        if self.proj.cyInstance.nb_dendrites() < self.proj.post.size:
            post_idx = np.where( np.array(self.proj.cyInstance.post_rank()) == self.post_rank )[0][0]
        else:
            post_idx = self.post_rank

        # Set default values for the additional variables
        extra_attributes = []
        for var in self.proj.synapse_type.description['parameters'] + self.proj.synapse_type.description['variables']:
            if not var['name'] in ['w', 'delay'] and  var['name'] in self.proj.synapse_type.description['local']:
                if isinstance(self.proj.init[var['name']], (int, float, bool)):
                    init = var['init']
                elif isinstance(self.proj.init[var], RandomDistribution):
                    init = self.proj.init[var].get_value()
                else:
                    init = self.proj.init[var['name']]
                extra_attributes.append(init)

        try:
            self.proj.cyInstance.add_single_synapse(
                post_idx, 
                rank, 
                w, 
                int(delay/ConfigManager().get('dt', self.proj.net_id)), 
                *extra_attributes
            )
        except Exception as e:
            Messages._print(e)
            Messages._error(f'Dendrite.create_synapse(): Could not add synapse of rank {rank} to the dendrite {self.post_rank}.')

    def create_synapses(self, ranks:list[int], weights:float|list[float]=None, delays: float|list[float]=None) -> None:
        """
        Creates a set of synapses for this dendrite with the given pre-synaptic neurons.

        The configuration key `'structural_plasticity'` must be set to `True` before `compile()` for this method to work. 

        ```python
        net = ann.Network()
        net.config(structural_plasticity=True)
        net.compile()

        try:
            proj.dendrite(10).create_synapses(ranks=[20, 30, 40], weights=0.1, delay=0.0)
        except Exception as e:
            print(e)
        ```

        If the synapses already exist, an error is thrown, so make sure to catch the exception.

        :param ranks: list of ranks for the pre-synaptic neurons.
        :param weights: list of synaptic weights (default: 0.0).
        :param delays: list of synaptic delays (default = dt).
        """
        if not ConfigManager().get('structural_plasticity', self.proj.net_id):
            Messages._error('"structural_plasticity" has not been set to True in setup(), can not add the synapses.')
            return

        # If not all neurons in the post-synaptic population receive connections
        # the post-rank diverge from the LIL index
        if self.proj.cyInstance.nb_dendrites() < self.proj.post.size:
            post_idx = np.where(self.proj.cyInstance.post_rank, self.post_rank)
        else:
            post_idx = self.post_rank

        # Process pre-synaptic ranks
        if isinstance(ranks, list):
            ranks = np.array(ranks)

        # Process weights
        if weights is None:
            # No user-defined init
            weights = np.array([0.0] * len(ranks))
        elif isinstance(weights, float):
            weights = np.array([weights] * len(ranks))
        elif isinstance(weights, list):
            # User provided a list
            weights = np.array(weights)

        # Process delays
        if delays is None:
            delays = np.array([0] * len(ranks))
        elif isinstance(delays, float):
            delays = np.array([delays] * len(ranks))
        # convert milliseconds -> steps
        delays = np.array(delays/ConfigManager().get('dt', self.proj.net_id), dtype=np.int32)

        # Sort ranks, weights, and delays by ascending ranks
        sorted_indices = np.argsort(ranks)
        ranks = ranks[sorted_indices]
        weights = weights[sorted_indices]
        delays = delays[sorted_indices]

        # Collect other attributes than w/delay
        extra_attributes = []
        for var in self.proj.synapse_type.description['parameters'] + self.proj.synapse_type.description['variables']:
            if not var['name'] in ['w', 'delay'] and var['name'] in self.proj.synapse_type.description['local']:
                if isinstance(self.proj.init[var['name']], (int, float, bool)):
                    init = [self.proj.init[var['name']]] * len(ranks)
                elif isinstance(self.proj.init[var['name']], RandomDistribution):
                    init = self.proj.init[var['name']].get_list_values(len(ranks))
                else:
                    raise AttributeError
                extra_attributes.append(np.array(init))

        # Update connectivity
        try:
            self.proj.cyInstance.add_multiple_synapses(
                post_idx, 
                ranks, 
                weights, 
                delays, 
                *extra_attributes
            )
        except Exception as e:
            Messages._print(e)

    def prune_synapse(self, rank:int) -> None:
        """
        Removes the synapse with the given pre-synaptic neuron from the dendrite.

        The configuration key `'structural_plasticity'` must be set to `True` before `compile()` for this method to work. 

        ```python
        net = ann.Network()
        net.config(structural_plasticity=True)
        net.compile()

        try:
            proj.dendrite(10).prune_synapse(rank=20)
        except Exception as e:
            print(e)
        ```

        If the synapse does not exist, an error is thrown, so make sure to catch the exception.

        :param rank: rank of the pre-synaptic neuron
        """
        if not ConfigManager().get('structural_plasticity', self.proj.net_id):
            Messages._error('"structural_plasticity" has not been set to True in setup(), can not remove the synapse.')
            return

        if not rank in self.pre_ranks:
            Messages._error('the synapse with the pre-synaptic neuron of rank ' + str(rank) + ' did not exist.')
            return

        self.proj.cyInstance.remove_single_synapse(self.post_rank, rank)

    def prune_synapses(self, ranks:list[int]):
        """
        Removes the synapses which belong to the provided pre-synaptic neurons from the dendrite.

        The configuration key `'structural_plasticity'` must be set to `True` before `compile()` for this method to work. 

        ```python
        net = ann.Network()
        net.config(structural_plasticity=True)
        net.compile()

        try:
            proj.dendrite(10).prune_synapses(ranks=[20, 30, 40])
        except Exception as e:
            print(e)
        ```

        If the synapses do not exist, an error is thrown, so make sure to catch the exception.

        :param ranks: list of ranks of the pre-synaptic neurons.
        """
        if not ConfigManager().get('structural_plasticity', self.proj.net_id):
            Messages._error('"structural_plasticity" has not been set to True in setup(), can not remove the synapses.')
            return

        for rank in ranks:
            self.prune_synapse(rank)

class IndividualSynapse :

    def __init__(self, dendrite, rank):
        self.dendrite = dendrite
        self.rank = rank

    def __getattr__(self, name):
        " Method called when accessing an attribute."
        if name in ['dendrite', 'attributes', 'rank']:
            return object.__getattribute__(self, name)

        elif name in self.dendrite.proj.synapse_type.description['attributes']:
            # Determine C++ data type
            ctype = None
            for var in self.dendrite.proj.synapse_type.description['variables']+self.dendrite.proj.synapse_type.description['parameters']:
                if var['name'] == name:
                    ctype = var['ctype']

            if name in self.dendrite.proj.synapse_type.description['local']:
                return getattr(self.dendrite.proj.cyInstance, "get_local_attribute_"+ctype)(name, self.dendrite.idx, self.rank)
            elif name in self.dendrite.proj.synapse_type.description['semiglobal']:
                return getattr(self.dendrite.proj.cyInstance, "get_semiglobal_attribute_"+ctype)(name, self.dendrite.idx)
            else:
                return getattr(self.dendrite.proj.cyInstance, "get_global_attribute_"+ctype)(name)

        else:
            return object.__getattribute__(self, name)

    def __setattr__(self, name, value):
        " Method called when setting an attribute."
        if name in ['dendrite', 'attributes', 'rank']:
            object.__setattr__(self, name, value)

        elif name in self.dendrite.proj.synapse_type.description['attributes']:
            # Determine C++ data type
            ctype = None
            for var in self.dendrite.proj.synapse_type.description['variables']+self.dendrite.proj.synapse_type.description['parameters']:
                if var['name'] == name:
                    ctype = var['ctype']

            if name in self.dendrite.proj.synapse_type.description['local']:
                getattr(self.dendrite.proj.cyInstance, "set_local_attribute_"+ctype)(name, self.dendrite.idx, self.rank, value)
            elif name in self.dendrite.proj.synapse_type.description['semiglobal']:
                getattr(self.dendrite.proj.cyInstance, "set_semiglobal_attribute_"+ctype)(name, self.dendrite.idx, value)
            else:
                getattr(self.dendrite.proj.cyInstance, "set_global_attribute_"+ctype)(name, value)

        else:
            object.__setattr__(self, name, value)



"""
Contains global available functions and state variables,
e.g., the config dictionary and holds a reference to 
network instances.

:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from typing import Any

from ANNarchy.intern.ConfigManagement import ConfigManager, _update_global_config
from ANNarchy.intern.NetworkManager import NetworkManager
from ANNarchy.intern.GlobalObjects import GlobalObjectManager
from ANNarchy.intern.Profiler import Profiler
from ANNarchy.intern import Messages

__all__ = [
    'clear',
    'check_profile_results',
    'reset',
    'magic_network',
    'get_population',
    'get_projection',
    'populations',
    'projections',
    'monitors',
    'add_function',
    'functions',
    'enable_learning',
    'disable_learning',
    'get_time',
    'set_time',
    'get_current_step',
    'set_current_step',
    'dt',
    'set_seed',
]

# Minimum number of neurons to apply OMP parallel regions
OMP_MIN_NB_NEURONS = 100

def clear(functions:bool=True, neurons:bool=True, synapses:bool=True):
    """
    Clears all variables (erasing already defined populations, projections, monitors), as if you had just imported ANNarchy.

    Useful when re-running Jupyter/IPython notebooks multiple times:

    ```python
    import ANNarchy as ann
    ann.clear()
    ```

    :param functions: if True (default), all functions defined with ``add_function`` are erased.
    :param neurons: if True (default), all neurons defined with ``Neuron`` are erased.
    :param synapses: if True (default), all synapses defined with ``Synapse`` are erased.
    """
    # Reset globally defined objects
    GlobalObjectManager().clear(
        functions=functions, 
        neurons=neurons, 
        synapses=synapses, 
    )

    # Remove the present profiler
    if Profiler().enabled:
        check_profile_results()
        Profiler().disable_profiling()

    # Reinitialize initial state
    NetworkManager().clear()

def check_profile_results():
    """
    If the user enabled profiling, we here check if we recorded some results.
    """
    if Profiler().enabled:
        Profiler().print_profile()

        Profiler().store_cpp_time_as_csv()

def reset(populations:bool=True, projections:bool=False, synapses:bool=False, monitors:bool=True, reseed_rng:bool=True, net_id:int=0):
    """
    Reinitialises the network to its state before the call to `compile()`. The network time will be set to 0ms.

    All monitors are emptied.

    :param populations: if True (default), the neural parameters and variables will be reset to their initial value.
    :param projections: if True, the synaptic parameters and variables (except the connections) will be reset (default=False).
    :param synapses: if True, the synaptic weights will be erased and recreated (default=False).
    :param monitors: if True, the monitors will be emptied and reset (default=True).
    ;param reseed_rng: if True, RNG generators will be reset using the stored seed (default=True).
    """
    NetworkManager().get_network(net_id=net_id).reset(populations=populations, projections=projections, synapses=synapses, monitors=monitors, reseed_rng=reseed_rng)

################################
## Accessing shadow network
################################
def magic_network() -> "Network":
    """
    Returns the magic network of ID 0.

    The magic network collects populations and projections created directly, like in the old (<5.0) API.

    ```python
    pop = ann.Population(1000, ann.Izhikevich)
    proj = ann.Projection(pop, pop, 'exc)

    net = magic_network()
    net.compile()
    ```
    """
    return NetworkManager().get_network(net_id=0)

def get_population(name:str, net_id:int=0) -> "Population":
    """
    Returns the population with the given name.

    :param name: name of the population.
    """
    for pop in NetworkManager().get_network(net_id=net_id).get_populations():
        if pop.name == name:
            return pop

    Messages._warning("get_population(): the population", name, "does not exist.")
    return None

def get_projection(name:str, net_id:int=0) -> "Projection":
    """
    Returns the projection with the given name.

    :param name: name of the projection.
    """
    for proj in NetworkManager().get_network(net_id=net_id).get_projections():
        if proj.name == name:
            return proj

    Messages._warning("get_projection(): the projection", name, "does not exist.")
    return None

def populations(net_id:int=0) -> list["Population"]:
    """
    Returns a list of all declared populations.

    :retruns: a list of all populations.
    """
    return NetworkManager().get_network(net_id=net_id).get_populations()

def projections(
        net_id:int=0, 
        post:"Population"=None, 
        pre:"Population"=None, 
        target:str=None, 
        suppress_error:bool=False) -> list["Projection"]:
    """
    Returns a list of all declared populations. 
    """
    return NetworkManager().get_network(net_id=net_id).get_projections()

def monitors(net_id:int=0, obj: Any=None) -> list["Monitor"]:
    """
    Returns a list of declared monitors. 
    
    By default, all monitors are returned.
    
    By setting *obj*, only monitors recording from this object, either *Population* or *Projection*, will be returned.
    """
    if obj is None:
        return NetworkManager().get_network(net_id=net_id).get_monitors()

    else:
        mon_list = []
        for monitor in NetworkManager().get_network(net_id=net_id).get_monitors():
            if monitor.object == obj:
                mon_list.append(monitor)

        return mon_list

################################
## Functions
################################
def add_function(function:str):
    """
    Defines a global function which can be used by all neurons and synapses.

    The function must have only one return value and use only the passed arguments.

    Examples of valid functions:

    ```python
    ann.add_function('logistic(x) = 1 / (1 + exp(-x))')

    ann.add_function('''
        piecewise(x, a, b) = if x < a:
                                a
                             else: 
                                if x > b :
                                    b
                                else:
                                    x
    ''')
    ```

    Please refer to the manual to know the allowed mathematical functions.

    :param function: (multi)string representing the function.
    """
    GlobalObjectManager().add_function(function)

def functions(name:str, network=None):
    """
    Allows to access a global function declared with ``add_function()`` and use it from Python using arrays **after compilation of the magic network**.

    The name of the function is not added to the global namespace to avoid overloading.
    
    ```python
    add_function("logistic(x) = 1. / (1. + exp(-x))") 

    magic_network().compile()  

    result = functions('logistic')([0., 1., 2., 3., 4.])
    ```

    Only lists or 1D Numpy arrays can be passed as arguments, not single values nor multidimensional arrays.

    When passing several arguments, make sure they have the same size.

    :param name: name of the function.
    """
    return GlobalObjectManager().functions(name, network)

################################
## Memory management
################################
def _bytes_human_readable(size_in_bytes):
    """ Transforms given size in GB/MB/KB or bytes dependent on the value. """
    if size_in_bytes > (1024*1024*1024):
        return "{:.2f} GB".format(float(size_in_bytes)/(1024.0*1024.0*1024.0))
    elif size_in_bytes > (1024*1024):
        return "{:.2f} MB".format(float(size_in_bytes)/(1024.0*1024.0))
    elif size_in_bytes > (1024):
        return "{:.2f} KB".format(float(size_in_bytes)/(1024.0))
    else:
        return str(size_in_bytes) + " bytes"

def _cpp_memory_footprint(net_id=0):
    """
    Print the C++ memory consumption for populations, projections on the console.

    :param net_id: net_id of the requested network.
    """
    NetworkManager().get_network(net_id=net_id)._cpp_memory_footprint()

def _python_current_max_rusage():
    """
    Prints the current max residen size for the current process and the children.
    """
    import resource
    size_kilobytes = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
    print(_bytes_human_readable(size_kilobytes*1024))

################################
## Learning flags
################################
def enable_learning(projections:list=None, period:list=None, offset:float=None, net_id:int=0):
    """
    Enables learning for all projections. Optionally *period* and *offset* can be changed for all projections.

    :param projections: the projections whose learning should be enabled. By default, all the existing projections are enabled.
    :param period: determines how often the synaptic variables will be updated.
    :param offset: determines the offset at which the synaptic variables will be updated relative to the current time.

    """
    if not projections:
        projections = NetworkManager().get_network(net_id=net_id).get_projections()
    for proj in projections:
        proj.enable_learning(period, offset)

def disable_learning(projections=None, net_id=0):
    """
    Disables learning for all projections.

    :param projections: the projections whose learning should be disabled. By default, all the existing projections are disabled.
    """
    if not projections:
        projections = NetworkManager().get_network(net_id=net_id).get_projections()
    for proj in projections:
        proj.disable_learning()

################################
## Time
################################
def get_time(net_id=0) -> float:
    "Returns the current time in ms."
    try:
        t = NetworkManager().get_network(net_id=net_id).instance.get_time() * ConfigManager().get('dt', net_id)
    except:
        t = 0.0
    return t

def set_time(t:float, net_id=0):
    """
    Sets the current time in ms.

    **Warning:** can be dangerous for some spiking models.
    """
    try:
        NetworkManager().get_network(net_id=net_id).instance.set_time(int(t / ConfigManager().get('dt', net_id)))
    except:
        Messages._warning('Time can only be set when the network is compiled.')

def get_current_step(net_id=0) -> int:
    "Returns the current simulation step."
    try:
        t = NetworkManager().get_network(net_id=net_id).instance.get_time()
    except:
        t = 0
    return t

def set_current_step(t:int, net_id=0):
    """
    Sets the current simulation step (integer).

    **Warning:** can be dangerous for some spiking models.
    """
    try:
        NetworkManager().get_network(net_id=net_id).instance.set_time(int(t))
    except:
        Messages._warning('Time can only be set when the network is compiled.')

def dt(net_id:int=0) -> float:
    "Returns the simulation step size `dt` used in the simulation."
    return ConfigManager().get('dt', net_id)

################################
## Seed
################################
def set_seed(seed:int, use_seed_seq:bool=True, net_id:int=0):
    """
    Sets the seed of the random number generators, both in ANNarchy.RandomDistributions and in the C++ library when it is created.

    Numpy still has to be seeded explicitly when using the default RNG, for example:

    ```python
    ann.set_seed(seed=42)
    rng = np.random.default_rng(seed=42)
    A = rng.uniform(0.0, 1.0, (10, 10))
    ```
    
    :param seed: integer value used to seed the C++ and Numpy RNG
    :param use_seed_seq: for openMP and parallel RNGs, we use either the STL SeedSequence (True, default) or a specialized implementation proposed by Melissa O'Neil (False, see _optimization_flags for more details).
    """

    if net_id == 0:
        _update_global_config('seed', seed)
        _update_global_config('use_seed_seq', use_seed_seq)
    
    try:
        if ConfigManager().get('disable_parallel_rng', net_id):
            NetworkManager().get_network(net_id=net_id).instance.set_seed(seed, 1, use_seed_seq)
        else:
            NetworkManager().get_network(net_id=net_id).instance.set_seed(seed, ConfigManager().get('num_threads', net_id), use_seed_seq)
    except:
        Messages._warning('The seed will only be set in the simulated network when it is compiled.')



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

import time

from ANNarchy.core.PopulationView import PopulationView
from ANNarchy.intern.Profiler import Profiler
from ANNarchy.intern.ConfigManagement import ConfigManager, _check_paradigm
from ANNarchy.intern.NetworkManager import NetworkManager
from ANNarchy.intern.GlobalObjects import GlobalObjectManager
from ANNarchy.intern import Messages
from ANNarchy.parser.Extraction import extract_functions

from ANNarchy.generator.NanoBind.Generator import NanoBindGenerator
from ANNarchy.generator.Monitor.MonitorGenerator import MonitorGenerator
from ANNarchy.generator.Population import SingleThreadGenerator, OpenMPGenerator, CUDAGenerator
from ANNarchy.generator.Projection import SingleThreadProjectionGenerator, OpenMPProjectionGenerator, CUDAProjectionGenerator
from ANNarchy.generator.Template.GlobalOperationTemplate import global_operation_templates_st, global_operation_templates_openmp, global_operation_templates_cuda
from ANNarchy.generator.Utils import tabify
from ANNarchy.generator.Template import BaseTemplate
from ANNarchy.generator import Profile

class CodeGenerator(object):
    """
    The CodeGenerator class is responsible to control the code
    generation process.

    Unil now, it implements the code generation for OpenMP
    (including sequential) and CUDA.The decision whether as
    OpenMP or sequential code is dependent on the number of
    threads.
    """
    def __init__(self, annarchy_dir, net_id, cuda_config):
        """
        Constructor initializes the PopulationGenerator and ProjectionGenerator
        class and stores the provided information for later use.

        Parameters:

            * *net_id*: unique id for the current network
            * *annarchy_dir*: unique target directory for the generated code
              files; they are stored in 'generate' sub-folder
            * *cuda_config*: configuration dict for cuda. check the method
              _cuda_kernel_config for more details.
        """
        self.net_id = net_id
        self._network = NetworkManager().get_network(self.net_id)
        self._annarchy_dir = annarchy_dir
        self._populations = self._network.get_populations() 
        self._projections = self._network.get_projections()
        self._cuda_config = cuda_config

        # Profiling is optional, but if either Global.config["profiling"] set to True
        # or --profile was added on command line.
        if ConfigManager().get('profiling', self.net_id):
            if ConfigManager().get('paradigm', self.net_id) == "openmp":
                self._profgen = Profile.CPP11Profile(self._annarchy_dir, net_id)
                self._profgen.generate()
            elif ConfigManager().get('paradigm', self.net_id) == "cuda":
                self._profgen = Profile.CUDAProfile(self._annarchy_dir, net_id)
                self._profgen.generate()
            else:
                Messages._error('No ProfileGenerator available for '
                              + ConfigManager().get('paradigm', self.net_id))
        else:
            self._profgen = None

        # Instantiate code generator based on the target platform
        if ConfigManager().get('paradigm', self.net_id) == "openmp":
            if ConfigManager().get('num_threads', self.net_id) == 1:
                self._popgen = SingleThreadGenerator(self._profgen, net_id)
                self._projgen = SingleThreadProjectionGenerator(self._profgen, net_id)
            else:
                self._popgen = OpenMPGenerator(self._profgen, net_id)
                self._projgen = OpenMPProjectionGenerator(self._profgen, net_id)
        elif ConfigManager().get('paradigm', self.net_id) == "cuda":
            self._popgen = CUDAGenerator(self._cuda_config['cuda_version'], self._profgen, net_id)
            self._projgen = CUDAProjectionGenerator(self._cuda_config['cuda_version'], self._profgen, net_id)
        else:
            Messages._error("No PopulationGenerator for " + ConfigManager().get('paradigm', self.net_id))

        # Py-extenstion and RecordGenerator are commonly defined
        self._nb_gen = NanoBindGenerator(annarchy_dir, net_id)
        self._recordgen = MonitorGenerator(annarchy_dir, net_id)

        # Target container for the generated code snippets
        self._pop_desc = []
        self._proj_desc = []

    def generate(self):
        """
        Generate code files and store them in target directory (located at
        self._annarchy_dir/generate). More detailed the following files are
        generated, by this class:

            * *ANNarchy.cpp*: main simulation loop, object instantiation
            * *ANNarchy.hpp*: collection of all objects, interface to Python
              extension
            * *ANNarchyCore.pyx*: Python extension file, gathering all
               functions/ objects, which should be accessible from Python
            * for each population a seperate header file, contain semantic
              logic of a population respectively neuron object (filename:
              pop<id>).
            * for each projection a seperate header file, contain semantic
              logic of a projection respectively synapse object (filename:
              proj<id>)
        """
        if Profiler().enabled:
            t0 = time.time()

        if ConfigManager().get('verbose', self.net_id):
            if ConfigManager().get('paradigm', self.net_id) == "openmp":
                if ConfigManager().get('num_threads', self.net_id) > 1:
                    Messages._print('\nGenerate code for OpenMP ...')
                else:
                    Messages._print('\nGenerate sequential code ...')
            elif ConfigManager().get('paradigm', self.net_id) == "cuda":
                print('\nGenerate CUDA code ...')
            else:
                raise NotImplementedError

        # Specific populations/projections have an overwritten _generate()
        # method which will populate the self._specific_template dictionary
        for pop in self._populations:
            pop._generate()
        for proj in self._projections:
            proj._generate()

        # Propagate the global operations needed by the projections to the
        # corresponding populations.
        self._propagate_global_ops()

        # Create all populations
        for pop in self._populations:
            self._pop_desc.append(self._popgen.header_struct(pop, self._annarchy_dir))

        # Create all projections
        for proj in self._projections:
            self._proj_desc.append(self._projgen.header_struct(proj, self._annarchy_dir))

        # where all source files should take place
        source_dest = self._annarchy_dir+'/generate/net'+str(self.net_id)+'/'

        # Generate header code for the analysed pops and projs
        if ConfigManager().get('paradigm', self.net_id) == "openmp":
            with open(source_dest+'ANNarchy.hpp', 'w') as ofile:
                ofile.write(self._generate_header())

        elif ConfigManager().get('paradigm', self.net_id) == "cuda":
            invoke_header, host_header = self._generate_header()
            with open(source_dest+'ANNarchyKernel.cuh', 'w') as ofile:
                ofile.write(invoke_header)
            with open(source_dest+'ANNarchy.hpp', 'w') as ofile:
                ofile.write(host_header)

        else:
            raise NotImplementedError

        # Generate monitor code for the analysed pops and projs
        self._recordgen.generate()

        # Generate cpp code for the analysed pops and projs
        if ConfigManager().get('paradigm', self.net_id) == "openmp":
            with open(source_dest+'ANNarchy.cpp', 'w') as ofile:
                ofile.write(self._generate_body())

        elif ConfigManager().get('paradigm', self.net_id) == "cuda":
            device_code, host_code = self._generate_body()
            with open(source_dest+'ANNarchy.cpp', 'w') as ofile:
                ofile.write(host_code)
            with open(source_dest+'ANNarchyKernel.cu', 'w') as ofile:
                ofile.write(device_code)

        else:
            raise NotImplementedError

        # Generate cython code for the analysed pops and projs
        with open(source_dest+'ANNarchyCore'+str(self.net_id)+'.cpp', 'w') as ofile:
            ofile.write(self._nb_gen.generate())

        self._generate_file_overview(source_dest)

        if Profiler().enabled:
            t1 = time.time()
            Profiler().add_entry(t0, t1, "generate", "compile")

    def _generate_file_overview(self, source_dest):
        """
        Generate a logfile, where we log which Population/Projection object is stored in
        which file.

        Parameters:

        * source_dest: path to folder where generated files are stored.
        """
        pop_desc = """pop%(id_pop)s, %(type_pop)s(name ='%(name_pop)s', neuron='%(neuron_type)s')\n"""
        proj_desc = """proj%(id_proj)s, %(type_proj)s(pre='%(pre_name)s', post='%(post_name)s', target='%(target)s', synapse='%(synapse_type)s', name='%(name)s') using connector: %(pattern)s \n"""

        # Equal to target path in CodeGenerator.generate()
        with open(source_dest+"codegen.log", 'w', encoding="utf-8") as ofile:
            ofile.write("Filename, Object Description\n")
            for pop in self._populations:
                pop_type = type(pop).__name__
                desc_dict = {
                    'id_pop': pop.id,
                    'name_pop': pop.name,
                    'neuron_type': pop.neuron_type.name,
                    'type_pop': pop_type
                }
                ofile.write(pop_desc % desc_dict)

            for proj in self._projections:
                proj_type = type(proj).__name__
                desc_dict = {
                    'id_proj': proj.id,
                    'type_proj': proj_type,
                    'pre_name': proj.pre.name,
                    'post_name': proj.post.name,
                    'target': proj.target,
                    'synapse_type': proj.synapse_type.name,
                    'name': proj.name
                }

                # In case of debug, we print the parameters otherwise not
                if ConfigManager().get('debug', self.net_id):
                    desc_dict.update({'pattern': proj.connector_description})
                else:
                    desc_dict.update({'pattern': proj.connector_description.split(',')[0]})

                ofile.write(proj_desc % desc_dict)

    def _propagate_global_ops(self):
        """
        The parser analyses the synapse and neuron definitions and
        store if global operations like min, max or mean are necessary.

        Furthermore for synapses accesses to population variales (e. g. pre.r)
        occure. In this case we need to generate special codes in the PopulationGenerator.
        """
        # Analyse the populations
        for pop in self._populations:
            pop.global_operations = pop.neuron_type.description['global_operations']
            pop.delayed_variables = []

        # Propagate the global operations from the projections to the populations
        for proj in self._projections:
            for op in proj.synapse_type.description['pre_global_operations']:
                if isinstance(proj.pre, PopulationView):
                    if not op in proj.pre.population.global_operations:
                        proj.pre.population.global_operations.append(op)
                else:
                    if not op in proj.pre.global_operations:
                        proj.pre.global_operations.append(op)

            for op in  proj.synapse_type.description['post_global_operations']:
                if isinstance(proj.post, PopulationView):
                    if not op in proj.post.population.global_operations:
                        proj.post.population.global_operations.append(op)
                else:
                    if not op in proj.post.global_operations:
                        proj.post.global_operations.append(op)

            if proj.max_delay > 1:
                for var in proj.synapse_type.description['dependencies']['pre']:
                    if isinstance(proj.pre, PopulationView):
                        proj.pre.population.delayed_variables.append(var)
                    else:
                        proj.pre.delayed_variables.append(var)

        # Make sure the operations are declared only once
        for pop in self._populations:
            pop.global_operations = [dict(y) for y in sorted(set(tuple(x.items())) for x in pop.global_operations)]
            pop.delayed_variables = sorted(list(set(pop.delayed_variables)))

    def _generate_header(self):
        """
        Generate the ANNarchy.hpp code. This header represents the interface to
        the Python extension and therefore includes all network objects.
        """
        # struct declaration for each population
        pop_struct = ""
        pop_ptr = ""

        for pop in self._pop_desc:
            pop_struct += pop['include']
            pop_ptr += pop['extern']

        # struct declaration for each projection
        proj_struct = ""
        proj_ptr = ""
        for proj in self._proj_desc:
            proj_struct += proj['include']
            proj_ptr += proj['extern']

        # Custom functions
        custom_func = self._header_custom_functions()

        # Custom constants
        custom_constant = self._header_custom_constants()

        # Final code
        header_code = ""
        if ConfigManager().get('paradigm', self.net_id) == "openmp":
            header_code = BaseTemplate.omp_header_template % {
                'float_prec': ConfigManager().get('precision', self.net_id),
                'pop_struct': pop_struct,
                'proj_struct': proj_struct,
                'pop_ptr': pop_ptr,
                'proj_ptr': proj_ptr,
                'custom_func': custom_func,
                'custom_constant': custom_constant,
                'built_in': BaseTemplate.built_in_functions + BaseTemplate.integer_power_cpu % {'float_prec': ConfigManager().get('precision', self.net_id)},
            }
            return header_code

        elif ConfigManager().get('paradigm', self.net_id) == "cuda":
            # kernel declaration
            invoke_kernel_def = ""
            for pop in self._pop_desc:
                invoke_kernel_def += pop['update_header']

            for proj in self._proj_desc:
                invoke_kernel_def += proj['psp_kernel_decl']
                invoke_kernel_def += proj['update_synapse_header']
                invoke_kernel_def += proj['postevent_header']

            glob_ops_header, _, _ = self._body_def_glops()
            invoke_kernel_def += glob_ops_header

            device_invoke_header = BaseTemplate.cuda_device_invoke_header % {
                'float_prec': ConfigManager().get('precision', self.net_id),
                'invoke_kernel_def': invoke_kernel_def
            }

            host_header_code = BaseTemplate.cuda_header_template % {
                'float_prec': ConfigManager().get('precision', self.net_id),
                'pop_struct': pop_struct,
                'proj_struct': proj_struct,
                'pop_ptr': pop_ptr,
                'proj_ptr': proj_ptr,
                'custom_func': custom_func,
                'built_in': BaseTemplate.built_in_functions,
                'custom_constant': custom_constant
            }
            return device_invoke_header, host_header_code
        
        else:
            raise NotImplementedError

    def _header_custom_functions(self):
        """
        Generate code for custom functions defined globally and are usable
        witihn neuron or synapse descriptions. These functions can only rely on
        provided arguments.
        """
        if GlobalObjectManager().number_functions() == 0:
            return ""

        # Attention CUDA: this definition will work only on host side.
        code = ""
        for _, func in GlobalObjectManager().get_functions():
            code += extract_functions(description=func, local_global=True, net_id=self.net_id)[0]['cpp'] + '\n'

        return code

    def _header_custom_constants(self):
        """
        Generate code for custom constants
        """
        network = NetworkManager().get_network(self.net_id)
        constants = network.get_constants()

        if len(constants) == 0:
            return ""

        code = ""
        for obj in constants:
            obj_str = {
                'name': obj.name,
                'float_prec': ConfigManager().get('precision', self.net_id)
            }
            if _check_paradigm("openmp", self.net_id):
                code += """
extern %(float_prec)s %(name)s;
void set_%(name)s(%(float_prec)s value);""" % obj_str
            elif _check_paradigm("cuda", self.net_id):
                code += """
void set_%(name)s(%(float_prec)s value);""" % obj_str
            else:
                raise NotImplementedError

        return code

    def _body_custom_constants(self):
        """
        Generate code for custom constants dependent on the target paradigm
        set in global settings.

        Returns (openMP):

        * decl_code: declarations in header file
        * init_code: initialization code

        Returns (CUDA):

        * device_decl_code: declarations in header file (device side)
        * host_init_code: initialization code (host side)

        """
        network = NetworkManager().get_network(self.net_id)
        constants = network.get_constants()

        if len(constants) == 0:
            return "", ""
        
        if _check_paradigm("openmp", self.net_id):

            decl_code = ""
            init_code = ""
            for obj in constants:
                obj_str = {
                    'name': obj.name,
                    'value': obj.value,
                    'float_prec': ConfigManager().get('precision', self.net_id)
                }
                decl_code += """
%(float_prec)s %(name)s;
void set_%(name)s(%(float_prec)s value){%(name)s = value;};""" % obj_str
                init_code += """
        %(name)s = 0.0;""" % obj_str

            return decl_code, init_code

        elif _check_paradigm("cuda", self.net_id):

            host_init_code = ""
            device_decl_code = ""
            for obj in constants:
                obj_str = {
                    'name': obj.name,
                    'value': obj.value,
                    'float_prec': ConfigManager().get('precision', self.net_id)
                }
                device_decl_code += """__device__ __constant__ %(float_prec)s %(name)s;
void set_%(name)s(%(float_prec)s value) {
    cudaError_t err = cudaMemcpyToSymbol(%(name)s, &value, sizeof(%(float_prec)s), 0, cudaMemcpyHostToDevice);
#ifdef _DEBUG
    std::cout << "set global constant %(name)s = " << value << std::endl;
    if ( err != cudaSuccess )
        std::cerr << cudaGetErrorString(err) << std::endl;
#endif
}""" % obj_str

                # TODO: is this really needed, it's overwritten anyways ?
                host_init_code += """
        set_%(name)s(0.0);""" % obj_str

            return device_decl_code, host_init_code

        else:
            raise NotImplementedError

    def _generate_body(self):
        """
        Generate the codes 'main' library file. The generated code
        will be used in different files, dependent on the chosen
        target platform:

        * openmp: ANNarchy.cpp
        * cuda: ANNarchyHost.cu and ANNarchyDevice.cu
        """
        # struct declaration for each population
        pop_ptr = ""
        for pop in self._pop_desc:
            pop_ptr += pop['instance']

        # struct declaration for each projection
        proj_ptr = ""
        for proj in self._proj_desc:
            proj_ptr += proj['instance']

        # Code for the global operations
        glop_definition = self._body_def_glops()
        update_globalops = ""
        for pop in self._pop_desc:
            if 'gops_update' in pop.keys():
                update_globalops += pop['gops_update']

        # Reset presynaptic sums
        reset_sums = self._body_resetcomputesum_pop()

        # Compute presynaptic sums
        compute_sums = ""
        # Sum over all synapses
        if _check_paradigm("openmp", self.net_id):
            for proj in self._proj_desc:
                compute_sums += proj["compute_psp"]

        # Update random distributions
        rd_update_code = ""
        for desc in self._pop_desc + self._proj_desc:
            if 'rng_update' in desc.keys():
                rd_update_code += desc['rng_update']

        # Equations for the neural variables
        update_neuron = ""
        for pop in self._pop_desc:
            if 'update' in pop.keys():
                update_neuron += pop['update']

        # Enque delayed outputs
        delay_code = ""
        for pop in self._pop_desc:
            if 'delay_update' in pop.keys():
                delay_code += pop['delay_update']

        # Equations for the synaptic variables
        update_synapse = ""
        for proj in self._proj_desc:
            if 'update' in proj.keys():
                update_synapse += proj['update']

        # Equations for the post-events
        post_event = ""
        for proj in self._proj_desc:
            if 'post_event' in proj.keys():
                post_event += proj['post_event']

        # Structural plasticity
        structural_plasticity, sp_spike_backward_view_update = self._body_structural_plasticity()

        # Early stopping
        run_until = self._body_run_until()

        #Profiling
        if self._profgen:
            prof_dict = self._profgen.generate_body_dict()
            prof_dict['prof_include'] = prof_dict['prof_include'].replace("extern ", "")
        else:
            prof_dict = Profile.ProfileGenerator(self._annarchy_dir, self.net_id).generate_body_dict()

        #
        # Generate the ANNarchy.cpp code, the corrsponding template differs
        # greatly. For further information take a look into the corresponding
        # branches.
        #
        if ConfigManager().get('paradigm', self.net_id) == "openmp":
            # custom constants
            custom_constant, _ = self._body_custom_constants()

            # code fields for openMP/single thread template
            base_dict = {
                'float_prec': ConfigManager().get('precision', self.net_id),
                'pop_ptr': pop_ptr,
                'proj_ptr': proj_ptr,
                'glops_def': glop_definition,
                'initialize': self._body_initialize(),
                'run_until': run_until,
                'compute_sums' : compute_sums,
                'reset_sums' : reset_sums,
                'update_neuron' : update_neuron,
                'update_globalops' : update_globalops,
                'update_synapse' : update_synapse,
                'random_dist_update' : rd_update_code,
                'delay_code' : delay_code,
                'post_event' : post_event,
                'structural_plasticity': structural_plasticity,
                'custom_constant': custom_constant,
                'sp_spike_backward_view_update': sp_spike_backward_view_update
            }

            # profiling
            base_dict.update(prof_dict)

            # complete code template
            if ConfigManager().get('num_threads', self.net_id) == 1:
                return BaseTemplate.st_body_template % base_dict
            else:
                return BaseTemplate.omp_body_template % base_dict

        elif ConfigManager().get('paradigm', self.net_id) == "cuda":
            # Implementation notice ( HD: 10. June, 2015 )
            #
            # The CUDA linking process is a big problem for object oriented approaches
            # and the seperation of implementation codes into several files. Even in the
            # current SDK 5.0 this problem is not fully solved. Linking is available, but
            # only for small, independent code pieces, by far not sufficient for full
            # object-oriented approaches ...
            #
            # For us, this currently have one consequence: we cannot completely seperate
            # the implementation of objects into several files. To hold a certain equality
            # between the structures of objects, I implemented the following workaround:
            #
            # We create the c-structs holding data fields and accessors as in OpenMP. We also
            # create the kernels, call entity in the corresponding generator objects, and
            # return the codes via the descriptor dictionary.
            #
            # This ensures a consistent interface in the generators and also in the generated
            # codes, but sometimes require additional overhead. Hopefully NVidia will improve
            # their linker in the next releases, so one could remove this overhead.
            psp_call = ""
            for proj in self._proj_desc:
                psp_call += proj['psp_host_call']

            # custom constants
            device_custom_constant, _ = self._body_custom_constants()

            # custom functions
            custom_func = ""
            for pop in self._pop_desc:
                custom_func += pop['custom_func']
            for proj in self._proj_desc:
                custom_func += proj['custom_func']
            for _, func in GlobalObjectManager().get_functions():
                custom_func += extract_functions(
                    description=func, 
                    local_global=True, 
                    net_id=self.net_id)[0]['cpp'].replace("inline", "__device__") + '\n'

            # pre-defined/common available kernel
            common_kernel = self._cuda_common_kernel(self._projections)

            pop_kernel = ""
            pop_invoke_kernel = ""
            for pop in self._pop_desc:
                pop_kernel += pop['update_body']
                pop_invoke_kernel += pop['update_invoke']

            pop_update_fr = ""
            for pop in self._pop_desc:
                pop_update_fr += pop['update_FR']

            psp_device_kernel = ""
            psp_invoke_kernel = ""
            for proj in self._proj_desc:
                psp_device_kernel += proj['psp_device_kernel']
                psp_invoke_kernel += proj['psp_invoke_kernel']

            delay_code = ""
            for pop in self._pop_desc:
                if 'update_delay' in pop.keys():
                    delay_code += pop['update_delay']

            syn_kernel = ""
            syn_invoke_kernel = ""
            for proj in self._proj_desc:
                syn_kernel += proj['update_synapse_body']
                syn_invoke_kernel += proj['update_synapse_invoke']

            syn_call = ""
            for proj in self._proj_desc:
                syn_call += proj['update_synapse_call']

            postevent_device_kernel = ""
            postevent_invoke_kernel = ""
            for proj in self._proj_desc:
                postevent_device_kernel += proj['postevent_body']
                postevent_invoke_kernel += proj['postevent_invoke']

            postevent_call = ""
            for proj in self._proj_desc:
                postevent_call += proj['postevent_call']

            clear_sums = self._body_resetcomputesum_pop()

            # global operations
            _, glob_ops_invoke, glob_ops_body = self._body_def_glops()

            # determine number of threads per kernel
            threads_per_kernel = self._cuda_kernel_config()

            # concurrent kernel execution
            stream_setup = self._cuda_stream_config()

            # memory transfers
            host_device_transfer, device_host_transfer = "", ""
            for pop in self._pop_desc:
                host_device_transfer += pop['host_to_device']
                # DtoH is performed only when an accessor is called.

            for proj in self._proj_desc:
                host_device_transfer += proj['host_to_device']
                # DtoH is performed only when an accessor is called.

            # Profiling
            if self._profgen:
                prof_dict = self._profgen.generate_body_dict()
                prof_dict['prof_include'] = prof_dict['prof_include'].replace("extern ", "")
            else:
                prof_dict = Profile.ProfileGenerator(self._annarchy_dir, self.net_id).generate_body_dict()

            device_code = BaseTemplate.cuda_device_kernel % {      # Target: ANNarchyKernel.cu
                'common_kernel': common_kernel,
                'pop_kernel': pop_kernel,
                'pop_invoke_kernel': pop_invoke_kernel,
                'psp_kernel': psp_device_kernel,
                'psp_invoke_kernel': psp_invoke_kernel,
                'syn_kernel': syn_kernel,
                'syn_invoke_kernel': syn_invoke_kernel,
                'glob_ops_kernel': glob_ops_body,
                'glob_ops_invoke_kernel': glob_ops_invoke,
                'postevent_kernel': postevent_device_kernel,
                'postevent_invoke_kernel': postevent_invoke_kernel,
                'custom_func': custom_func,
                'custom_constant': device_custom_constant,
                'built_in': BaseTemplate.built_in_functions + BaseTemplate.integer_power_cuda % {'float_prec': ConfigManager().get('precision', self.net_id)},
                'float_prec': ConfigManager().get('precision', self.net_id)
            }

            base_dict = {
                # network definitions
                'float_prec': ConfigManager().get('precision', self.net_id),
                'pop_ptr': pop_ptr,
                'proj_ptr': proj_ptr,
                'run_until': run_until,
                'clear_sums': clear_sums,
                'compute_sums' : psp_call,
                'update_neuron' : update_neuron,
                'update_FR': pop_update_fr,
                'update_globalops' : update_globalops,
                'update_synapse' : syn_call,
                'post_event': postevent_call,
                'delay_code': delay_code,
                'initialize' : self._body_initialize(),
                'structural_plasticity': structural_plasticity,

                # cuda host specific
                'stream_setup': stream_setup,
                'host_device_transfer': host_device_transfer,
                'device_host_transfer': device_host_transfer,
                'kernel_config': threads_per_kernel,
                'sp_spike_backward_view_update': ""
            }
            base_dict.update(prof_dict)
            host_code = BaseTemplate.cuda_host_body_template % base_dict    # Target: ANNarchy.cpp

            return device_code, host_code
        else:
            raise NotImplementedError

    def _body_initialize(self):
        """
        Define codes for the method initialize(), comprising of population and projection
        initializations, optionally profiling class.
        """
        profiling_init = "" if not ConfigManager().get('profiling', self.net_id) else self._profgen.generate_init_network()

        # Initialize populations
        population_init = "    // Initialize populations\n"
        for pop in self._pop_desc:
            population_init += pop['init']

        # Initialize projections
        projection_init = "    // Initialize projections\n"
        for proj in self._proj_desc:
            projection_init += proj['init']

        # Initialize custom constants
        if ConfigManager().get('paradigm', self.net_id) == "openmp":
            # Custom  constants
            _, custom_constant = self._body_custom_constants()

            init_tpl = BaseTemplate.omp_initialize_template
        elif ConfigManager().get('paradigm', self.net_id) == "cuda":
            # Custom  constants
            _, custom_constant = self._body_custom_constants()

            init_tpl = BaseTemplate.cuda_initialize_template
        else:
            raise NotImplementedError

        return init_tpl % {
            'prof_init': profiling_init,
            'pop_init': population_init,
            'proj_init': projection_init,
            'custom_constant': custom_constant
        }

    def _body_resetcomputesum_pop(self):
        """
        Rate-coded neurons sum up the received inputs in temporary variables.
        They need to be cleared in each time step.
        """
        code = ""
        for pop in self._populations:
            if pop.neuron_type.type == 'rate':
                code += self._popgen.reset_computesum(pop)

        return code

    def _body_structural_plasticity(self):
        """
        Call of pruning or creating methods if necessary.

        Returns two strings:
            * call statements called within singleStep()
            * call statements called at begin of simulation loop
        """
        # Pruning if any
        pruning = ""
        creating = ""
        rebuild_in_cpp = ""
        rebuild_out_cpp = ""

        if ConfigManager().get('structural_plasticity', self.net_id):
            for proj in self._projections:
                rebuild_needed = False
                if 'pruning' in proj.synapse_type.description.keys():
                    pruning += tabify("proj%(id)s->pruning();\n" % {'id': proj.id}, 1)
                    rebuild_needed = True
                if 'creating' in proj.synapse_type.description.keys():
                    creating += tabify("proj%(id)s->creating();\n" % {'id': proj.id}, 1)
                    rebuild_needed = True
                # we only check those projections which are possibly modified
                if rebuild_needed and proj.synapse_type.type == 'spike':
                    rebuild_in_cpp += tabify("proj%(id)s->check_and_rebuild_inverse_connectivity();\n" % {'id': proj.id}, 1)

                # we don't know which projection the user modifies, so we need to check all
                if proj.synapse_type.type == 'spike':
                    rebuild_out_cpp += tabify("proj%(id)s->check_and_rebuild_inverse_connectivity();\n" % {'id': proj.id}, 1)

        return creating + pruning + rebuild_in_cpp, rebuild_out_cpp

    def _body_def_glops(self):
        """
        Dependent on the used global operations we add pre-defined templates
        to the ANNarchy body file.

        Return:

            dependent on the used paradigm we return one string (single thread, OpenMP)
            or tuple(string, string) (CUDA).
        """
        ops = []
        for pop in self._populations:
            for op in pop.global_operations:
                ops.append(op['function'])

        # no global operations
        if ops == []:
            if _check_paradigm("openmp", self.net_id):
                return ""
            elif _check_paradigm("cuda", self.net_id):
                return "", "", ""
            else:
                raise NotImplementedError("CodeGenerator._body_def_glops(): no implementation for "+ConfigManager().get('paradigm', self.net_id))

        type_def = {
            'type': ConfigManager().get('precision', self.net_id)
        }

        # the computation kernel depends on the paradigm
        if _check_paradigm("openmp", self.net_id):
            if ConfigManager().get('num_threads', self.net_id) == 1:
                global_op_template = global_operation_templates_st
            else:
                global_op_template = global_operation_templates_openmp

            code = ""
            for op in sorted(list(set(ops))):
                code += global_op_template[op] % type_def

            return code

        elif _check_paradigm("cuda", self.net_id):
            header = ""
            invoke = ""
            body = ""

            for op in sorted(list(set(ops))):
                header += global_operation_templates_cuda[op]['header'] % type_def
                invoke += global_operation_templates_cuda[op]['invoke'] % type_def
                body += global_operation_templates_cuda[op]['body'] % type_def

            return header, invoke, body
        else:
            raise NotImplementedError("CodeGenerator._body_def_glops(): no implementation for "+ConfigManager().get('paradigm', self.net_id))

    def _body_run_until(self):
        """
        Generate the code for conditioned stop of simulation
        """
        tpl = BaseTemplate.omp_run_until_template

        # Check if it is useful to generate anything at all
        for pop in self._populations:
            if pop.stop_condition:
                break
        else:
            # No stop conditions were detected
            return tpl['default']

        # a condition has been defined, so we generate corresponding code
        cond_code = ""
        for pop in self._populations:
            if pop.stop_condition:
                cond_code += tpl['single_pop'] % {'id': pop.id}

        return tpl['body'] % {'run_until': cond_code}

#######################################################################
############## CUDA specific ##########################################
#######################################################################
    def _cuda_kernel_config(self):
        """
        Each GPU kernel requires a launch configuration, established in
        the ANNarchyHost.cu code. Until ANNarchy 4.7 we always defined
        the configuration as pre-processor symbol. In context of variations
        required by different formats, we changed the strategy and this
        symbols are only used IF the user wants to overwrite something.

        * 192 threads for psp and synapse update
        * guessed amount of threads for neurons, based on population size
          (see _guess_pop_kernel_config)

        Notice:

            Only related to the CUDA implementation
        """
        if self._cuda_config is None:
            return ""

        from math import ceil

        # Population config adjust neuron_update
        configuration = "// Populations\n"
        for pop in self._populations:
            if pop in self._cuda_config.keys():
                if 'num_threads' in self._cuda_config[pop].keys():
                    num_threads = self._cuda_config[pop]['num_threads']
                    num_blocks = int(ceil(float(pop.size)/float(num_threads)))

                if 'num_blocks' in self._cuda_config[pop].keys():
                    num_blocks = self._cuda_config[pop]['num_blocks']

                cfg = """#define __pop%(id)s_tpb__ %(nr)s
#define __pop%(id)s_nb__ %(nb)s
"""
                configuration += cfg % {
                    'id': pop.id,
                    'nr': num_threads,
                    'nb': num_blocks
                }

                if ConfigManager().get('verbose', self.net_id):
                    Messages._print('population', pop.id, ' - kernel config: (', num_blocks, ',', num_threads, ')')

        # Projection config - adjust psp, synapse_local_update, synapse_global_update
        configuration += "\n// Projections\n"
        for proj in self._projections:
            if proj in self._cuda_config.keys():
                if 'num_threads' in self._cuda_config[proj].keys():
                    num_threads = self._cuda_config[proj]['num_threads']
                if 'num_blocks' in self._cuda_config[proj].keys():
                    num_blocks = self._cuda_config[proj]['num_blocks']

                cfg = """#define __proj%(id_proj)s_%(target)s_tpb__ %(nr)s
#define __proj%(id_proj)s_%(target)s_nb__ %(nb)s
"""

                # proj.target can hold a single or multiple targets. We use
                # one configuration for all but need to define single names anyways
                target_list = proj.target if isinstance(proj.target, list) else [proj.target]
                for target in target_list:
                    configuration += cfg % {
                        'id_proj': proj.id,
                        'target': target,
                        'nr': num_threads,
                        'nb': num_blocks
                    }

                    if ConfigManager().get('verbose', self.net_id):
                        Messages._print('projection', proj.id, 'with target', target, ' - kernel config: (', num_blocks, ',', num_threads, ')')

        return configuration

    def _cuda_stream_config(self):
        """
        With Fermi Nvidia introduced multiple streams respectively concurrent
        kernel execution (requires device with compute compability > 2.x).

        Notice:

            Only related to the CUDA implementation
        """
        if self._cuda_config is None:
            pop_assign = "    // populations\n"
            proj_assign = "    // projections\n"
            max_number_streams = 0
        else:
            # TODO: maybe this should be a parameter too? As one could schedule multiple objects
            #       in one stream, the maximum number is not exploited
            max_number_streams = max(len(self._populations), len(self._projections))

            # HD:
            # the try-except blocks here are a REALLY lazy method.
            # TODO: it should be implemented more carefully in future
            pop_assign = "    // populations\n"
            for pop in self._populations:
                try:
                    sid = self._cuda_config[pop]['stream']
                    pop_assign += """    pop%(pid)s->stream = streams[%(sid)s];
""" % {'pid': pop.id, 'sid': sid}
                except KeyError:
                    # default stream, if either no cuda_config at all or
                    # the population is not configured by user
                    pop_assign += """    pop%(pid)s->stream = 0;
""" % {'pid': pop.id}

            proj_assign = "    // projections\n"
            for proj in self._projections:
                try:
                    sid = self._cuda_config[proj]['stream']
                    proj_assign += """    proj%(pid)s->stream = streams[%(sid)s];
""" % {'pid': proj.id, 'sid': sid}
                except KeyError:
                    # default stream, if either no cuda_config at all or
                    # the projection is not configured by user
                    proj_assign += """    proj%(pid)s->stream = 0;
""" % {'pid': proj.id}

        # Write config
        stream_config = BaseTemplate.cuda_stream_setup % {
            'nbStreams': max_number_streams,
            'pop_assign': pop_assign,
            'proj_assign': proj_assign
        }

        return stream_config

    def _guess_pop_kernel_config(self, pop):
        """
        Instead of a fixed amount of threads for each kernel, we try
        to guess a good configuration based on the population size.
        """
        from math import log

        max_tpb = 512
        warp_size = 32

        num_neur = pop.size / 2 # at least 2 iterations per thread
        guess = warp_size       # smallest block is 1 warp

        # Simplest case: we have more neurons than
        # available threads per block
        if num_neur > max_tpb:
            guess = max_tpb

        # check which is the closest possible thread amount
        pow_of_2 = [2**x for x in range(int(log(warp_size, 2)), int(log(max_tpb, 2))+1)]
        for i in range(len(pow_of_2)):
            if pow_of_2[i] < num_neur:
                continue
            else:
                guess = pow_of_2[i]
                break

        return guess

    def _guess_proj_kernel_config(self, proj):
        """
        Instead of a fixed amount of threads for each kernel, we try
        to guess a good configuration based on the pre-synaptic population size.
        """
        from math import log

        max_tpb = 512
        warp_size = 32

        num_neur = proj.pre.size / 4 # at least 1/4 of the neurons are connected
        guess = warp_size       # smallest block is 1 warp

        # Simplest case: we have more neurons than
        # available threads per block
        if num_neur > max_tpb:
            guess = max_tpb

        # check which is the closest possible thread amount
        pow_of_2 = [2**x for x in range(int(log(warp_size, 2)), int(log(max_tpb, 2))+1)]
        for i in range(len(pow_of_2)):
            if pow_of_2[i] < num_neur:
                continue
            else:
                guess = pow_of_2[i]
                break

        if ConfigManager().get('verbose', self.net_id):
            Messages._print('projection', proj.id, ' - kernel size:', guess)

        return guess

    def _cuda_common_kernel(self, projections):
        """
        Some sparse matrix formats require additional functions. Which we need to
        define only once.
        """
        fmts = []

        for proj in projections:
            fmts.append(proj._storage_format)

        fmts = list(set(fmts))

        code = ""
        # TODO: generalize!
        if "csr" in fmts:
            from ANNarchy.generator.Projection.CUDA.CSR import additional_global_functions
            code += additional_global_functions
        elif "csr_vector" in fmts:
            from ANNarchy.generator.Projection.CUDA.CSR_Vector import additional_global_functions
            code += additional_global_functions

        return code
